from Bio import SeqIO
import os
from collections import OrderedDict
import argparse
import networkx as nx
import vamb
import json

from collections.abc import Sequence
from typing import cast


def main(
    contig_names: Sequence[str],
    contig_lengths: Sequence[int],
    cluster_contigs: dict[str, set[str]],
    bin_path: dict[str, str],
    path_ripped: str,
    path_clusters_not_ripped: str,
) -> dict[str, str]:
    cluster_contigs_original = cluster_contigs.copy()

    contig_length = get_contig_length(contig_names, contig_lengths)
    cluster_length = get_cluster_length(cluster_contigs, contig_length)
    graph_clusters = get_graph_clusters(cluster_contigs, contig_length, cluster_length)
    (
        cluster_contigs,
        graph_clusters,
        cluster_length,
        clusters_changed_but_not_intersecting_contigs,
    ) = remove_meaningless_edges_from_pairs(
        graph_clusters, cluster_length, cluster_contigs, contig_length
    )

    (
        cluster_contigs,
        graph_clusters,
        cluster_length,
        clusters_changed_but_not_intersecting_contigs_2,
    ) = make_all_components_pair(
        graph_clusters, cluster_contigs, cluster_length, contig_length
    )

    ripped_clusters_set = create_ripped_clusters_and_write_ripped_bins(
        graph_clusters, cluster_contigs, bin_path, path_ripped
    )

    clusters_changed_but_not_intersecting_contigs_total = (
        clusters_changed_but_not_intersecting_contigs.copy()
    )
    clusters_changed_but_not_intersecting_contigs_total.update(
        clusters_changed_but_not_intersecting_contigs_2
    )

    (
        bin_path,
        cluster_contigs_not_ripped,
    ) = find_remaining_clusters_ripped_and_write_ripped_bins(
        cluster_contigs_original,
        clusters_changed_but_not_intersecting_contigs_total,
        bin_path,
        path_ripped,
    )

    # Remove the clusters that have been ripped because of removing the intersections
    for cluster in ripped_clusters_set:
        if cluster in cluster_contigs_not_ripped.keys():
            print("%s cluster removed from the set since it was ripped" % (cluster))
            del cluster_contigs_not_ripped[cluster]

    with open(path_clusters_not_ripped, "w") as file:
        print(vamb.vambtools.CLUSTERS_HEADER, file=file)
        for cluster, contigs_nr in cluster_contigs_not_ripped.items():
            contigs_o = cluster_contigs_original[cluster]
            if contigs_o != contigs_nr:
                raise ValueError(
                    "Clusters that have been ripped remain in the non ripped set of clusters"
                )
            for contig in contigs_nr:
                print(cluster, contig, sep="\t", file=file)
            file.flush()
    return bin_path


def get_contig_length(names: Sequence[str], lengths: Sequence[int]) -> dict[str, int]:
    return dict(zip(names, lengths))


def get_cluster_length(
    cluster_contigs: dict[str, set[str]], contig_length: dict[str, int]
) -> OrderedDict[str, int]:
    cluster_length: OrderedDict[str, int] = OrderedDict()
    for cluster_name, contigs in cluster_contigs.items():
        cluster_length[cluster_name] = 0
        for contig in contigs:
            cluster_length[cluster_name] += contig_length[contig]
    return cluster_length


def get_graph_clusters(
    cluster_contigs: dict[str, set[str]],
    contig_length: dict[str, int],
    cluster_length: OrderedDict[str, int],
):
    """Generate a graph where each cluster is a node and each edge connecting 2 nodes is
    created if clusters share some contigs, edges are weighted by the intersection length
    """
    graph_clusters = nx.Graph()
    for i, cluster_i in enumerate(cluster_contigs.keys()):
        for j, cluster_j in enumerate(cluster_contigs.keys()):
            if j <= i:
                j += 1
                continue

            cluster_i_contigs, cluster_j_contigs = (
                cluster_contigs[cluster_i],
                cluster_contigs[cluster_j],
            )
            cluster_i_j_contigs_intersection = cluster_i_contigs.intersection(
                cluster_j_contigs
            )

            if len(cluster_i_j_contigs_intersection) > 0:
                cluster_i_length, cluster_j_length = (
                    cluster_length[cluster_i],
                    cluster_length[cluster_j],
                )
                cluster_i_j_contigs_intersection_length = sum(
                    (
                        contig_length[contig_k]
                        for contig_k in cluster_i_j_contigs_intersection
                    )
                )
                intersection_length_ratio = (
                    cluster_i_j_contigs_intersection_length
                    / min(cluster_i_length, cluster_j_length)
                )
                graph_clusters.add_node(cluster_i)
                graph_clusters.add_node(cluster_j)
                graph_clusters.add_edge(
                    cluster_i, cluster_j, weight=intersection_length_ratio
                )
    return graph_clusters


def remove_meaningless_edges_from_pairs(
    graph_clusters: nx.Graph,
    cluster_length: OrderedDict[str, int],
    cluster_contigs: dict[str, set[str]],
    contig_length: dict[str, int],
    weight_thr: float = 0.001,
):
    """Iterate over all the components of the graph and remove components composed by 2 node/clusters
    according the following criteria:
        - For edges with weight smaller than threshold, just remove edge and remove intersecting contigs from
          larger cluster
    """
    components: list[set[str]] = list()

    for component in nx.connected_components(graph_clusters):
        components.append(component)

    clusters_changed_but_not_intersecting_contigs: dict[str, set[str]] = dict()
    for component in components:
        if len(component) == 2:
            cl_i = component.pop()
            cl_j = component - {cl_i}
            assert len(cl_j) == 1
            cl_j = cl_j.pop()
            cl_i_cl_j_edge_weight: float = graph_clusters.get_edge_data(cl_i, cl_j)[
                "weight"
            ]
            if cl_i_cl_j_edge_weight > weight_thr:
                pass
            else:
                cluster_updated = move_intersection_to_smaller_cluster(
                    graph_clusters,
                    cl_i,
                    cl_j,
                    cluster_contigs,
                    cluster_length,
                    contig_length,
                )
                print("Cluster ripped because of a meaningless edge ", cluster_updated)
                clusters_changed_but_not_intersecting_contigs[cluster_updated] = (
                    cluster_contigs[cluster_updated]
                )

    components: list[set[str]] = list()
    for component in nx.connected_components(graph_clusters):
        components.append(component)

    for component in components:
        if len(component) == 1:
            cl_i = component.pop()
            graph_clusters.remove_node(cl_i)

    print("Meaningless edges removed from pairs")

    return (
        cluster_contigs,
        graph_clusters,
        cluster_length,
        clusters_changed_but_not_intersecting_contigs,
    )


def move_intersection_to_smaller_cluster(
    graph_bins: nx.Graph,
    cl_i: str,
    cl_j: str,
    cluster_contigs: dict[str, set[str]],
    cluster_length: OrderedDict[str, int],
    contig_length: dict[str, int],
) -> str:
    """Move the intersecting contigs to the shortest cluster"""
    cl_i_contigs, cl_j_contigs = cluster_contigs[cl_i], cluster_contigs[cl_j]
    cl_i_cl_j_intersection_contigs = cl_i_contigs.intersection(cl_j_contigs)
    cl_i_length, cl_j_length = cluster_length[cl_i], cluster_length[cl_j]

    if cl_i_length < cl_j_length:
        cluster_contigs[cl_j] = cl_j_contigs - cl_i_cl_j_intersection_contigs
        cluster_length[cl_j] = sum(
            [contig_length[contig_i] for contig_i in cluster_contigs[cl_j]]
        )
        cluster_updated = cl_j
    else:
        cluster_contigs[cl_i] = cl_i_contigs - cl_i_cl_j_intersection_contigs
        cluster_length[cl_i] = sum(
            [contig_length[contig_i] for contig_i in cluster_contigs[cl_i]]
        )
        cluster_updated = cl_i
    graph_bins.remove_edge(cl_i, cl_j)

    return cluster_updated


def make_all_components_pair(
    graph_clusters: nx.Graph,
    cluster_contigs: dict[str, set[str]],
    cluster_length: OrderedDict[str, int],
    contig_length: dict[str, int],
):
    """Iterate over all graph components that contain more than 2 nodes/clusters and break them down to
    2 nodes/clusters components according the following criteria:
        - Remove the weaker edge by removing the intersecting contigs from the larger cluster
    """

    components: list[set[str]] = list()
    for component in nx.connected_components(graph_clusters):
        components.append(component)

    clusters_changed_but_not_intersecting_contigs: dict[str, set[str]] = dict()
    for component in components:
        if len(component) > 2:
            weight_edges: dict[float, set[str]] = dict()
            for cl_i in component:
                for cl_j in component - {cl_i}:
                    if graph_clusters.has_edge(cl_i, cl_j):
                        cl_i_cl_j_edge_weight: float = graph_clusters.get_edge_data(
                            cl_i, cl_j
                        )["weight"]
                        # if the weight as key in the dict and is mapping to different clusters (nodes)
                        # then add a insignificant value to the weight so we can add the weight with
                        # different clusters (nodes)
                        if (
                            cl_i_cl_j_edge_weight in weight_edges.keys()
                            and weight_edges[cl_i_cl_j_edge_weight] != {cl_i, cl_j}
                        ):
                            cl_i_cl_j_edge_weight += 1 * 10**-30
                        weight_edges[cl_i_cl_j_edge_weight] = {cl_i, cl_j}

            dsc_sorted_weights: list[float] = sorted(list(weight_edges.keys()))

            i = 0
            component_len = len(component)

            print(
                dsc_sorted_weights,
                len(dsc_sorted_weights),
                len(component),
                weight_edges.keys(),
            )
            while component_len > 2:
                weight_i = dsc_sorted_weights[i]
                cl_i, cl_j = weight_edges[weight_i]
                print(weight_i, cl_i, cl_j)
                cluster_updated = move_intersection_to_smaller_cluster(
                    graph_clusters,
                    cl_i,
                    cl_j,
                    cluster_contigs,
                    cluster_length,
                    contig_length,
                )
                print("Cluster ripped because of a pairing component ", cluster_updated)
                clusters_changed_but_not_intersecting_contigs[cluster_updated] = (
                    cluster_contigs[cluster_updated]
                )
                component_len = max(
                    [
                        len(nx.node_connected_component(graph_clusters, node_i))
                        for node_i in component
                    ]
                )

                i += 1

    components: list[set[str]] = list()
    for component in nx.connected_components(graph_clusters):
        components.append(component)

    for component in components:
        assert len(component) < 3

        if len(component) == 1:
            cl_i = component.pop()
            graph_clusters.remove_node(cl_i)
    print("All components are reduced to 2 nodes per component")
    return (
        cluster_contigs,
        graph_clusters,
        cluster_length,
        clusters_changed_but_not_intersecting_contigs,
    )


def create_ripped_clusters_and_write_ripped_bins(
    graph_clusters: nx.Graph,
    cluster_contigs: dict[str, set[str]],
    bin_path: dict[str, str],
    path_ripped: str,
) -> set[str]:
    """Given the components of the graph, write ripped bins where each bin misses the intersection with the other bin"""
    ripped_clusters_set: set[str] = set()
    try:
        os.mkdir(path_ripped)
    except FileExistsError:
        pass

    for component in nx.connected_components(graph_clusters):
        cl_i = next(iter(component))
        cl_j = component - {cl_i}
        assert len(cl_j) == 1
        cl_j = next(iter(cl_j))
        cl_i_contigs, cl_j_contigs = cluster_contigs[cl_i], cluster_contigs[cl_j]
        cluster_i_j_contigs_intersection = cl_i_contigs.intersection(cl_j_contigs)
        print(
            "Cluster %s and cluster %s share %i contigs "
            % (cl_i, cl_j, len(cluster_i_j_contigs_intersection))
        )

        bin_i_path = bin_path[cl_i + ".fna"]
        bin_j_path = bin_path[cl_j + ".fna"]

        bin_i_ripped_path = os.path.join(path_ripped, cl_i + "--" + cl_j + ".fna")
        bin_j_ripped_path = os.path.join(path_ripped, cl_j + "--" + cl_i + ".fna")
        print(bin_i_path, bin_i_ripped_path)
        write_ripped_bin(
            cl_i_contigs - cluster_i_j_contigs_intersection,
            bin_i_path,
            bin_i_ripped_path,
        )

        print(bin_j_path, bin_j_ripped_path)
        write_ripped_bin(
            cl_j_contigs - cluster_i_j_contigs_intersection,
            bin_j_path,
            bin_j_ripped_path,
        )

        ripped_clusters_set.add(cl_i)
        ripped_clusters_set.add(cl_j)

    if len(os.listdir(path_ripped)) == 0:
        os.rmdir(path_ripped)
    print("Ripped bins created from components")
    return ripped_clusters_set


def write_ripped_bin(
    contigs_cluster_ripped, bin_i_path: str, bin_i_ripped_path: str
) -> None:
    """Write ripped bin"""
    contig_sequences = []
    for contig in SeqIO.parse(bin_i_path, "fasta"):
        if contig.id in contigs_cluster_ripped:
            contig_sequences.append(contig)
    with open(bin_i_ripped_path, "w") as output_handle:
        SeqIO.write(contig_sequences, output_handle, "fasta")


def find_remaining_clusters_ripped_and_write_ripped_bins(
    cluster_contigs_original: dict[str, set[str]],
    clusters_changed_but_not_intersecting_contigs_total: dict[str, set[str]],
    bin_path: dict[str, str],
    path_ripped: str,
):
    """Compare cluster_contigs_ripped_only_by_meaningless_edges_or_not_ripped_at_all with cluster_contigs
    to find the clusters that remain different, so we can re evaluate them with checkM2, those clusters
    are the product of the meaningless ripping or the parition of n>2 components to generate
    all components n<=2"""
    cluster_contigs_original_copy = cluster_contigs_original.copy()

    try:
        os.mkdir(path_ripped)
    except FileExistsError:
        pass

    for (
        cluster,
        contigs_r,
    ) in clusters_changed_but_not_intersecting_contigs_total.items():
        contigs_o = cluster_contigs_original[cluster]

        assert contigs_r != contigs_o
        assert len(contigs_r) < len(contigs_o)
        bin_o_path = bin_path[cluster + ".fna"]
        bin_r_path = os.path.join(path_ripped, cluster + ".fna")
        write_ripped_bin(contigs_r, bin_o_path, bin_r_path)
        bin_path[cluster + ".fna"] = bin_r_path
        del cluster_contigs_original_copy[cluster]

    return bin_path, cluster_contigs_original_copy


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", type=str, help="path run generated by AVAMB")
    parser.add_argument("--ci", type=str, help="path dereplicated clusters file")
    parser.add_argument(
        "--co", type=str, help="path dereplicated clusters without ripped clusters"
    )
    parser.add_argument("-c", type=str, help="path to composition")
    parser.add_argument("--bp_d", type=str, help="bin_path dictionary path")
    parser.add_argument("--br", type=str, help="path ripped bins")
    parser.add_argument("--bin_separator", type=str, help="path ripped bins")
    parser.add_argument(
        "--log_nc_ripped_bins",
        type=str,
        help="path to og file that contains the # of ripped bins",
    )

    opt = parser.parse_args()

    path_run = (
        opt.r
    )  # '/Users/mbv396/aamb/vamb_aamb_on_almeida_hmp2/runs/vamb_aamb_on_almeida/'

    clusters_path = (
        opt.ci
    )  # os.path.join(path_run,'aae_l_y_vamb_manual_drep_85_10_clusters.tsv')
    clusters_not_ripped_path = (
        opt.co
    )  # os.path.join(path_run,'aae_l_y_vamb_manual_drep_85_10_not_ripped_clusters.tsv')

    with open(clusters_path) as file:
        cluster_contigs = vamb.vambtools.read_clusters(file)

    comp = vamb.parsecontigs.Composition.load(opt.c)
    contig_lengths = cast(Sequence[int], comp.metadata.lengths)
    contig_names = cast(Sequence[str], comp.metadata.identifiers)
    del comp  # free memory

    path_ripped = cast(str, opt.br)

    with open(opt.bp_d) as f:
        bin_path = json.load(f)

    bin_separator = opt.bin_separator

    bin_path_ = main(
        contig_names,
        contig_lengths,
        cluster_contigs,
        bin_path,
        path_ripped,
        clusters_not_ripped_path,
    )

    with open(opt.bp_d, "w") as f:
        json.dump(bin_path_, f)

    n_bins_ripped_in_folder = 0
    for _file in os.listdir(path_ripped):
        if ".fna" in _file:
            n_bins_ripped_in_folder += 1

    with open(opt.log_nc_ripped_bins, "w") as file:
        file.write(str(n_bins_ripped_in_folder))



import numpy as np
import os
from typing import cast


def get_cluster_score_bin_path(
    path_checkm_all: str, path_bins: str, bins: set[str]
) -> tuple[dict[str, tuple[float, float]], dict[str, str]]:
    """Given CheckM has been run for all samples, create 2 dictionaries:
    - {bin:path_bin}
    - {bin:[completeness, contamination]}"""
    cluster_score: dict[str, tuple[float, float]] = dict()
    bin_path: dict[str, str] = dict()
    for sample in os.listdir(path_checkm_all):
        path_quality_s = os.path.join(path_checkm_all, sample, "quality_report.tsv")
        c_com_con = np.loadtxt(
            path_quality_s,
            delimiter="\t",
            skiprows=1,
            usecols=(0, 1, 2),
            dtype=str,
            ndmin=2,
        )

        for row in c_com_con:
            cluster, com, con = row
            cluster = cast(str, cluster)
            com, con = float(com), float(con)
            bin_name = cluster + ".fna"
            if bin_name in bins:
                cluster_score[cluster] = (com, con)
                bin_path[cluster + ".fna"] = os.path.join(
                    path_bins, sample, cluster + ".fna"
                )
    return cluster_score, bin_path


def update_cluster_score_bin_path(
    path_checkm_ripped: str, cluster_score: dict[str, tuple[float, float]]
) -> dict[str, tuple[float, float]]:
    c_com_con = np.loadtxt(
        path_checkm_ripped,
        delimiter="\t",
        skiprows=1,
        usecols=(0, 1, 2),
        dtype=str,
        ndmin=2,
    )
    for row in c_com_con:
        cluster, com, con = row
        if "--" in cluster:
            continue
        com, con = float(com), float(con)
        print(cluster, "scores were", cluster_score[cluster])

        cluster_score[cluster] = (com, con)
        print("and now are", cluster_score[cluster])
    return cluster_score



import io
import os
import unittest
import tempfile
import gzip
import bz2
import lzma
import random
import itertools
import numpy as np
import string
import torch
import pathlib
import shutil

import vamb
from vamb.vambtools import BinSplitter
import testtools

PARENTDIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATADIR = os.path.join(PARENTDIR, "test", "data")


class TestReader(unittest.TestCase):
    def setUp(self):
        # Do not create files if they exist
        if hasattr(self, "files"):
            return None

        self.files = []
        with open(os.path.join(DATADIR, "fasta.fna"), "rb") as file:
            self.lines = list(file)

        for f in [gzip.compress, bz2.compress, lzma.compress, lambda x: x]:
            file = tempfile.NamedTemporaryFile()
            file.write(f(b"".join(self.lines)))
            file.flush()
            self.files.append(file)

        return None

    def test_reader(self):
        for file in self.files:
            with vamb.vambtools.Reader(file.name) as reader:
                self.assertEqual(self.lines, list(reader))


class TestPushArray(unittest.TestCase):
    def test_append(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=3)
        values = [1.0, 1, -3, 55.3]
        for i in values:
            arr.append(i)

        self.assertEqual(len(arr), 4)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))
        arr.clear()
        self.assertEqual(len(arr), 0)

    def test_extend(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        values = [1.0, 1, -3, 55.3, 0, 53, -1.2]
        arr.extend(values)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))

        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        for i in range(1, 10000, 100):
            arr.extend(range(i, i + 100))
        self.assertTrue(
            np.all(arr.take() == np.array(range(1, 10001), dtype=np.float32))
        )

    def test_clear(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        values = [1.0, 1, -3, 55.3, 0, 53, -1.2]
        arr.extend(values)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))
        arr.clear()
        self.assertTrue(len(arr.take()) == 0)


class TestFASTAEntry(unittest.TestCase):
    def test_init(self):
        passing = [b"0", b"foo", b"FOO123_@@^", b"%%;;/=~~"]
        failing = [
            b"",  # empty
            b">",  # empty
            b">foo",  # with >
            b"=",  # cannot start with =
            b"*",  # cannot start with *
            b"\xff",  # out of range
            b"))",
            b" ",  # zero-length identifier
        ]

        for i in passing:
            x = vamb.vambtools.FastaEntry(i, bytearray(b"TAG"))
            self.assertTrue(i.decode(), x.header)

        for i in failing:
            with self.assertRaises(ValueError):
                vamb.vambtools.FastaEntry(i, bytearray(b"TAG"))

        # Various
        with self.assertRaises(ValueError):
            vamb.vambtools.FastaEntry(b"foo", bytearray(b"ATGCpN"))

    def test_masking(self):
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"foo", bytearray(b"TaGkmYnAC")).sequence,
            bytearray(b"TaGkmYnAC"),
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"foo", bytearray()).sequence, bytearray()
        )

    def test_various(self):
        # Length
        self.assertEqual(len(vamb.vambtools.FastaEntry(b"x", bytearray(b"TAGCA"))), 5)
        self.assertEqual(len(vamb.vambtools.FastaEntry(b"x", bytearray())), 0)
        self.assertEqual(
            len(vamb.vambtools.FastaEntry(b"x", bytearray(b"TGTAmnyAncC"))), 11
        )

        # format
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"x", bytearray(b"TAGCA")).format(), ">x\nTAGCA"
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"yz", bytearray()).format(), ">yz\n"
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"1_2", bytearray(b"TGTAmnyAncC")).format(),
            ">1_2\nTGTAmnyAncC",
        )

    def test_kmercounts(self):
        seq = vamb.vambtools.FastaEntry(b"X", bytearray(b"TTAyCAAnGAC"))

        with self.assertRaises(ValueError):
            seq.kmercounts(0)

        with self.assertRaises(ValueError):
            seq.kmercounts(13)

        self.assertTrue(np.all(seq.kmercounts(1) == np.array([4, 2, 1, 2])))

        self.assertTrue(
            np.all(
                seq.kmercounts(2)
                == np.array(
                    [
                        1,
                        1,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        1,
                    ]
                )
            )
        )

    def test_random_kmercounts(self):
        indexof = {
            "".join(ncs): idx
            for (idx, ncs) in enumerate(itertools.product("ACGT", repeat=4))
        }
        seq = testtools.make_randseq(random.Random(), 900, 1100)
        sequence = seq.sequence.decode()
        manual_counts = np.zeros(256, dtype=int)
        for i in range(len(sequence) - 3):
            ind = indexof.get(sequence[i : i + 4].upper())
            if ind is not None:
                manual_counts[ind] += 1

        automatic = seq.kmercounts(4)
        self.assertTrue(np.all(manual_counts == automatic))

    def test_rename(self):
        seq = vamb.vambtools.FastaEntry(b"foo", bytearray(b"TaGkmYnAC"))

        # Does not error
        seq.rename(b"identifier\t desc")
        self.assertEqual(seq.identifier, "identifier")
        self.assertEqual(seq.description, "\t desc")

        seq.rename(b"newname")
        self.assertEqual(seq.identifier, "newname")
        self.assertEqual(seq.description, "")

        # Errors
        with self.assertRaises(ValueError):
            seq.rename(b"\tabc def")

        with self.assertRaises(ValueError):
            seq.rename(b"=123")

        with self.assertRaises(ValueError):
            seq.rename(b"")

        with self.assertRaises(ValueError):
            seq.rename(b"\xff")


class TestFASTAReader(unittest.TestCase):
    def test_bad_files(self):
        # First non-comment line must be header
        with self.assertRaises(ValueError):
            data = b"#foo\n#bar\n  \n>foo\nTAG".splitlines()
            list(vamb.vambtools.byte_iterfasta(data))

        # String input
        with self.assertRaises(TypeError):
            data = ">abc\nTAG\na\nAC".splitlines()
            list(vamb.vambtools.byte_iterfasta(data))  # type:ignore

    # Various correct formats
    def test_good_files(self):
        # Empty file
        data = b"".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(0, len(records))

        # Only comments
        data = b"#bar\n#foo\n#".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(0, len(records))

        # A few sequences
        data = b"#bar\n#foo\n>ab\nTA\nT\n \t\nA\nNN\n>foo\nCyAmmkg\n>bar\n".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(3, len(records))
        self.assertEqual(records[0].sequence, bytearray(b"TATANN"))
        self.assertEqual(records[1].sequence, bytearray(b"CyAmmkg"))
        self.assertEqual(records[2].sequence, bytearray(b""))

        # Comment inside seqs
        data = b"#bar\n>foo\nTAG\n##xxx\n>@@AA\nAAA".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(records[0].sequence, bytearray(b"TAG"))


class TestZscore(unittest.TestCase):
    arr = np.array([[1, 2, 2.5], [2, 4, 3], [0.9, 3.1, 2.8]])
    zscored = np.array(
        [
            [-1.44059316, -0.3865006, 0.14054567],
            [-0.3865006, 1.7216845, 0.66759195],
            [-1.54600241, 0.77300121, 0.45677344],
        ]
    )

    def almost_similar(self, x, y):
        self.assertTrue(np.all(np.abs(x - y) < 1e-6))

    def test_simple(self):
        self.almost_similar(vamb.vambtools.zscore(self.arr), self.zscored)

    def test_axes(self):
        self.almost_similar(
            vamb.vambtools.zscore(self.arr, axis=0),
            np.array(
                [
                    [-0.60404045, -1.26346568, -1.29777137],
                    [1.40942772, 1.18195176, 1.13554995],
                    [-0.80538727, 0.08151391, 0.16222142],
                ]
            ),
        )

        self.almost_similar(
            vamb.vambtools.zscore(self.arr, axis=1),
            np.array(
                [
                    [-1.33630621, 0.26726124, 1.06904497],
                    [-1.22474487, 1.22474487, 0.0],
                    [-1.40299112, 0.85548239, 0.54750873],
                ]
            ),
        )

    def test_axis_bounds(self):
        with self.assertRaises(np.exceptions.AxisError):
            vamb.vambtools.zscore(self.arr, axis=-1)

        with self.assertRaises(np.exceptions.AxisError):
            vamb.vambtools.zscore(self.arr, axis=2)

    def test_integer(self):
        with self.assertRaises(TypeError):
            vamb.vambtools.zscore(np.array([1, 2, 3]), inplace=True)

    def test_novar(self):
        arr = np.array([4, 4, 4])
        self.almost_similar(vamb.vambtools.zscore(arr), np.array([0, 0, 0]))

    def test_inplace(self):
        vamb.vambtools.zscore(self.arr, inplace=True)
        self.almost_similar(self.arr, self.zscored)


class TestInplaceMaskArray(unittest.TestCase):
    def almost_similar_np(self, x, y):
        self.assertTrue(np.all(np.abs(x - y) < 1e-6))

    def almost_similar_torch(self, x, y):
        self.assertTrue(torch.all(torch.abs(x - y) < 1e-6))

    def test_numpy(self):
        arr = np.random.random((10, 3)).astype(np.float32)
        mask = np.array([0, 1, 1, 1, 1, 0, 0, 0, 0, 0]).astype(np.uint8)
        arr2 = arr[[bool(i) for i in mask]]
        vamb.vambtools.numpy_inplace_maskarray(arr, mask)
        self.almost_similar_np(arr, arr2)

        # arr is now too short after being masked
        with self.assertRaises(ValueError):
            vamb.vambtools.numpy_inplace_maskarray(arr, mask)

    def test_torch(self):
        arr = torch.rand(10, 3)
        mask = torch.tensor([0, 1, 1, 1, 1, 0, 0, 0, 0, 0], dtype=bool)
        arr2 = arr[[bool(i) for i in mask]]
        vamb.vambtools.torch_inplace_maskarray(arr, mask)
        self.almost_similar_torch(arr, arr2)

        # arr is now too short after being masked
        with self.assertRaises(ValueError):
            vamb.vambtools.torch_inplace_maskarray(arr, mask)


class TestHashRefNames(unittest.TestCase):
    def test_refhash(self):
        names = ["foo", "9", "eleven", "a"]
        b1 = vamb.vambtools.RefHasher.hash_refnames(names)

        # Test that hashing them all at once is the same as hashing them one at a time
        hasher = vamb.vambtools.RefHasher()
        hasher.add_refname(names[0])
        hasher.add_refname(names[1])
        for j in names[2:]:
            hasher.add_refname(j)
        b7 = hasher.digest()

        names[1] = names[1] + "x"
        b2 = vamb.vambtools.RefHasher.hash_refnames(names)
        names[1] = names[1][:-1] + " \t"  # it strips whitespace off right end
        b3 = vamb.vambtools.RefHasher.hash_refnames(names)
        names = names[::-1]
        b4 = vamb.vambtools.RefHasher.hash_refnames(names)

        names = (i + "   " for i in names[::-1])
        b5 = vamb.vambtools.RefHasher.hash_refnames(names)
        b6 = vamb.vambtools.RefHasher.hash_refnames(names)  # now empty generator

        self.assertNotEqual(b1, b2)
        self.assertEqual(b1, b3)
        self.assertNotEqual(b1, b4)
        self.assertEqual(b1, b5)
        self.assertNotEqual(b1, b6)
        self.assertEqual(b1, b7)


class TestBinSplit(unittest.TestCase):
    before = [("bin1", ["s1-c1", "s1-c2", "s8-c1", "s9-c11"]), ("bin2", ["s12-c0"])]

    after = [
        ("s1-bin1", {"s1-c1", "s1-c2"}),
        ("s8-bin1", {"s8-c1"}),
        ("s9-bin1", {"s9-c11"}),
        ("s12-bin2", {"s12-c0"}),
    ]

    def test_inert(self):
        self.assertEqual(
            BinSplitter("").splitter, BinSplitter.inert_splitter().splitter
        )

    def test_split(self):
        self.assertEqual(list(BinSplitter("-").binsplit(self.before)), self.after)

    def test_badsep(self):
        with self.assertRaises(KeyError):
            list(BinSplitter("2").binsplit(self.before))

    def test_badtype(self):
        with self.assertRaises(Exception):
            list(BinSplitter("x").binsplit([(1, [2])]))  # type:ignore

    def test_nosplit(self):
        self.assertEqual(
            list(BinSplitter("").binsplit(self.before)),
            [(k, set(s)) for (k, s) in self.before],
        )

    def test_initialize(self):
        # Nothing happens to an inert splitter
        b = BinSplitter.inert_splitter()
        s = b.splitter
        b.initialize([""])
        self.assertEqual(s, b.splitter)

        b = BinSplitter("X")
        with self.assertRaises(ValueError):
            b.initialize(["AXC", "S1C2"])

        b = BinSplitter(None)
        b.initialize(["S1C2", "ABC"])
        self.assertEqual(b.splitter, None)

        b = BinSplitter(None)
        b.initialize(["S1C2", "KMCPLK"])

        b = BinSplitter("XYZ")
        b.initialize(["ABXYZCD", "KLMXYZA"])


class TestConcatenateFasta(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        rng = random.Random(19)
        ios = [io.BytesIO(), io.BytesIO()]
        for i in ios:
            for _ in range(4):
                fasta = testtools.make_randseq(rng, 10, 200)
                i.write(fasta.format().encode())
                i.write(b"\n")

        cls.file_1 = ios[0]
        cls.file_2 = ios[1]

    def setUp(self):
        self.file_1.seek(0)
        self.file_2.seek(0)

    def test_no_rename(self):
        out = io.StringIO()
        vamb.vambtools.concatenate_fasta_ios(
            out, [self.file_1, self.file_2], minlength=10, rename=False
        )
        out.seek(0)
        data = out.read()
        self.file_1.seek(0)
        self.file_2.seek(0)
        data_1 = self.file_1.read().decode()
        data_2 = self.file_2.read().decode()
        self.assertEqual(data, data_1 + data_2)

    def test_rename(self):
        out = io.StringIO()
        vamb.vambtools.concatenate_fasta_ios(
            out, [self.file_1, self.file_2], minlength=10, rename=True
        )
        out.seek(0)
        out_bytes = io.BytesIO(out.read().encode())
        records_out = list(vamb.vambtools.byte_iterfasta(out_bytes))
        self.file_1.seek(0)
        self.file_2.seek(0)
        records_1 = list(vamb.vambtools.byte_iterfasta(self.file_1))
        records_2 = list(vamb.vambtools.byte_iterfasta(self.file_2))
        records_in = records_1 + records_2
        self.assertEqual(
            [i.sequence for i in records_out], [i.sequence for i in records_in]
        )
        self.assertEqual(len(set(i.identifier for i in records_out)), len(records_out))

        # Check that sequences present in each file now starts with S1 and S2
        identifier_by_seq_hash = dict()
        for record in records_out:
            identifier_by_seq_hash[hash(bytes(record.sequence))] = record.identifier

        for prefix, records in [("S1", records_1), ("S2", records_2)]:
            for record in records:
                identifier = identifier_by_seq_hash[hash(bytes(record.sequence))]
                self.assertTrue(identifier.startswith(prefix))


class TestWriteClusters(unittest.TestCase):
    test_clusters = [("C1", set("abc")), ("C2", set("hkjlmn")), ("C3", set("x"))]
    io = io.StringIO()

    def setUp(self):
        self.io.truncate(0)
        self.io.seek(0)

    def linesof(self, string: str):
        return list(filter(lambda x: not x.startswith("#"), string.splitlines()))

    def conforms(self, str, clusters):
        lines = self.linesof(str)
        self.assertEqual(lines[0], vamb.vambtools.CLUSTERS_HEADER)
        self.assertEqual(len(lines) - 1, sum(len(v) for (_, v) in clusters))
        allcontigs = set()
        printed = set()
        printed_names = set()
        read = set()
        read_names = set()

        for k, v in clusters:
            allcontigs.update(v)

        for line in lines[1:]:
            name, _, contig = line.partition("\t")
            printed.add(contig)
            printed_names.add(name)

        for k, v in vamb.vambtools.read_clusters(io.StringIO(str)).items():
            read.update(v)
            read_names.add(k)

        self.assertEqual(allcontigs, printed)
        self.assertEqual(allcontigs, read)
        self.assertEqual(read_names, printed_names)

    def test_not_writable(self):
        buf = io.BufferedReader(io.BytesIO(b""))  # type:ignore
        with self.assertRaises(ValueError):
            vamb.vambtools.write_clusters(buf, self.test_clusters)  # type:ignore

    def test_normal(self):
        vamb.vambtools.write_clusters(self.io, self.test_clusters)
        self.conforms(self.io.getvalue(), self.test_clusters)

    def test_max_clusters(self):
        vamb.vambtools.write_clusters(self.io, self.test_clusters[:2])
        lines = self.linesof(self.io.getvalue())
        self.assertEqual(len(lines), 10)
        self.conforms(self.io.getvalue(), self.test_clusters[:2])


class TestWriteBins(unittest.TestCase):
    file = io.BytesIO()
    N_BINS = 10
    dir = pathlib.Path(
        os.path.join(
            tempfile.gettempdir(),
            "".join(random.choices(string.ascii_letters + string.digits, k=10)),
        )
    )

    @classmethod
    def setUpClass(cls):
        bins: dict[str, set[str]] = dict()
        seqs: dict[str, vamb.vambtools.FastaEntry] = dict()
        for _ in range(cls.N_BINS):
            binname = "".join(random.choices(string.ascii_letters, k=12))
            bins[binname] = set()
            for _ in range(random.randrange(3, 7)):
                seq = testtools.make_randseq(random.Random(), 100, 250)
                seqs[seq.identifier] = seq
                bins[binname].add(seq.identifier)
                cls.file.write(seq.format().encode())
                cls.file.write(b"\n")

        cls.bins = bins
        cls.seqs = seqs

    def setUp(self):
        self.file.seek(0)

    def tearDown(self):
        try:
            shutil.rmtree(self.dir)
        except FileNotFoundError:
            pass

    def test_bad_params(self):
        # Too many bins for maxbins
        with self.assertRaises(ValueError):
            vamb.vambtools.write_bins(
                self.dir, self.bins, self.file, maxbins=self.N_BINS - 1
            )

        # Parent does not exist
        with self.assertRaises(NotADirectoryError):
            vamb.vambtools.write_bins(
                pathlib.Path("svogew/foo"),
                self.bins,
                self.file,
                maxbins=self.N_BINS + 1,
            )

        # Target is an existing file
        with self.assertRaises(FileExistsError):
            with tempfile.NamedTemporaryFile() as file:
                vamb.vambtools.write_bins(
                    pathlib.Path(file.name),
                    self.bins,
                    self.file,
                    maxbins=self.N_BINS + 1,
                )

        # One contig missing from fasta dict
        with self.assertRaises(IndexError):
            bins = {k: v.copy() for k, v in self.bins.items()}
            next(iter(bins.values())).add("a_new_bin_which_does_not_exist")
            vamb.vambtools.write_bins(
                self.dir, bins, self.file, maxbins=self.N_BINS + 1
            )

    def test_round_trip(self):
        with tempfile.TemporaryDirectory() as dir:
            vamb.vambtools.write_bins(
                pathlib.Path(dir),
                self.bins,
                self.file,
                maxbins=self.N_BINS,
            )

            reconstructed_bins: dict[str, set[str]] = dict()
            for filename in os.listdir(dir):
                with open(os.path.join(dir, filename), "rb") as file:
                    entries = list(vamb.vambtools.byte_iterfasta(file))
                    binname = filename[:-4]
                    reconstructed_bins[binname] = set()
                    for entry in entries:
                        reconstructed_bins[binname].add(entry.identifier)

        # Same bins
        self.assertEqual(len(self.bins), len(reconstructed_bins))
        self.assertEqual(
            sum(map(len, self.bins.values())),
            sum(map(len, reconstructed_bins.values())),
        )
        self.assertEqual(self.bins, reconstructed_bins)



import unittest
import numpy as np
import tempfile
import vamb


class TestDataLoader(unittest.TestCase):
    def test_permute_indices(self):
        indices = vamb.semisupervised_encode.permute_indices(10, 25, seed=1)
        set_10 = set(range(10))
        self.assertTrue(len(indices) == 25)
        self.assertTrue(set(indices[:10]) == set_10)
        self.assertTrue(set(indices[10:20]) == set_10)
        self.assertTrue(set(indices[20:]).issubset(set_10))


class TestVAEVAE(unittest.TestCase):
    N_contigs = 111
    tnfs = np.random.random((N_contigs, 103)).astype(np.float32)
    rpkms = np.random.random((N_contigs, 14)).astype(np.float32)
    domain = "d_Bacteria"
    phyla = ["f_1", "f_2", "f_3"]
    classes = {
        "f_1": ["c_11", "c_21", "c_31"],
        "f_2": ["c_12", "c_22", "c_32"],
        "f_3": ["c_13", "c_23", "c_33"],
    }
    lengths = np.random.randint(2000, 5000, size=N_contigs)

    def make_random_annotation(self):
        phylum = np.random.choice(self.phyla, 1)[0]
        clas = np.random.choice(self.classes[phylum], 1)[0]
        if np.random.random() <= 0.2:
            return vamb.taxonomy.ContigTaxonomy.from_semicolon_sep(
                ";".join([self.domain])
            )
        if 0.2 < np.random.random() <= 0.5:
            return vamb.taxonomy.ContigTaxonomy.from_semicolon_sep(
                ";".join([self.domain, phylum])
            )
        return vamb.taxonomy.ContigTaxonomy.from_semicolon_sep(
            ";".join([self.domain, phylum, clas])
        )

    def make_random_annotations(self):
        return [self.make_random_annotation() for _ in range(self.N_contigs)]

    def test_make_graph(self):
        annotations = self.make_random_annotations()
        nodes, ind_nodes, table_parent = vamb.taxvamb_encode.make_graph(annotations)
        print(nodes, ind_nodes, table_parent)
        self.assertTrue(
            set(nodes).issubset(
                set(
                    [
                        "root",
                        "d_Bacteria",
                        "f_1",
                        "f_2",
                        "f_3",
                        "c_11",
                        "c_21",
                        "c_31",
                        "c_12",
                        "c_22",
                        "c_32",
                        "c_13",
                        "c_23",
                        "c_33",
                    ]
                )
            )
        )
        for p, cls in self.classes.items():
            for c in cls:
                for f in self.phyla:
                    # Since the taxonomy is generated randomly, we can't guarantee
                    # that each run will have all the clades.
                    if any(i not in ind_nodes for i in (p, c, f)):
                        continue
                    self.assertTrue(ind_nodes.get(f, -666) < ind_nodes.get(c, 666))
                    self.assertTrue(table_parent[ind_nodes[f]] == 1)
                    self.assertTrue(table_parent[ind_nodes[c]] == ind_nodes[p])

    def test_encoding(self):
        nlatent = 10
        batchsize = 10
        nepochs = 2
        annotations = self.make_random_annotations()
        nodes, ind_nodes, table_parent = vamb.taxvamb_encode.make_graph(annotations)

        classes_order = np.array([a.ranks[-1] for a in annotations])
        targets = np.array([ind_nodes[i] for i in classes_order])

        vae = vamb.taxvamb_encode.VAEVAEHLoss(
            self.rpkms.shape[1],
            len(nodes),
            nodes,
            table_parent,
            nlatent=nlatent,
            cuda=False,
        )

        dataloader_vamb = vamb.encode.make_dataloader(
            self.rpkms,
            self.tnfs,
            self.lengths,
            batchsize=batchsize,
            cuda=False,
        )
        dataloader_joint = vamb.taxvamb_encode.make_dataloader_concat_hloss(
            self.rpkms,
            self.tnfs,
            self.lengths,
            targets,
            len(nodes),
            table_parent,
            batchsize=batchsize,
            cuda=False,
        )
        dataloader_labels = vamb.taxvamb_encode.make_dataloader_labels_hloss(
            self.rpkms,
            self.tnfs,
            self.lengths,
            targets,
            len(nodes),
            table_parent,
            batchsize=batchsize,
            cuda=False,
        )

        shapes = (self.rpkms.shape[1], 103, 1, len(nodes))
        dataloader = vamb.taxvamb_encode.make_dataloader_semisupervised_hloss(
            dataloader_joint,
            dataloader_vamb,
            dataloader_labels,
            len(nodes),
            table_parent,
            shapes,
            666,
            batchsize=batchsize,
            cuda=False,
        )
        with tempfile.TemporaryFile() as modelfile:
            vae.trainmodel(
                dataloader,
                nepochs=nepochs,
                modelfile=modelfile,
                batchsteps=[],
            )

        latent_both = vae.VAEJoint.encode(dataloader_joint)
        self.assertEqual(latent_both.dtype, np.float32)
        self.assertEqual(latent_both.shape, (len(self.rpkms), nlatent))



import numpy as np
import argparse
import vamb
from pathlib import Path


def create_abundances(
    abundances: list[Path], mask_refhash: Path, min_id: float, outfile: Path
):
    """Merge the abundances to a single Abundance object and save it"""
    refhash = np.load(mask_refhash)["refhash"]

    n_samples = len(abundances)
    first = vamb.vambtools.read_npz(abundances[0])
    print(len(first), n_samples)
    print(first.shape)
    matrix = np.empty((len(first), n_samples), dtype=np.float32)
    matrix[:, 0] = first
    for i, path in enumerate(abundances[1:]):
        matrix[:, i + 1] = vamb.vambtools.read_npz(path)
    abundance = vamb.parsebam.Abundance(
        matrix, [str(i) for i in abundances], min_id, refhash
    )
    abundance.save(outfile)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--msk", type=Path, help="mask refhash")
    parser.add_argument("--ab", type=Path, nargs="+", help=" abundancaes list of files")
    parser.add_argument("--min_id", type=float, help="min identity for alignment")
    parser.add_argument("--out", type=Path, help="abundances outfile")

    opt = parser.parse_args()

    create_abundances(opt.ab, opt.msk, opt.min_id, opt.out)



import vamb
import argparse
import shutil
import os
import json

from typing import Optional


def main(
    cluster_scores: dict[str, tuple[float, float]],
    cluster_contigs: dict[str, set[str]],
    bin_separator: Optional[str],
    path_nc_bins_folder: str,
    path_bins_folder: str,
    path_nc_clusters: str,
    min_comp: float = 0.9,
    max_cont: float = 0.05,
):
    cluster_sample = get_cluster_sample(cluster_contigs, bin_separator)
    nc_cluster_scores = get_nc_cluster_scores(
        cluster_scores, cluster_sample, min_comp, max_cont
    )
    create_nc_sample_folders(nc_cluster_scores, cluster_sample, path_nc_bins_folder)
    write_nc_bins_from_mdrep_clusters(
        nc_cluster_scores, cluster_sample, path_nc_bins_folder, path_bins_folder
    )
    write_quality_report(nc_cluster_scores, path_nc_bins_folder)
    write_final_nc_clusters(nc_cluster_scores, cluster_contigs, path_nc_clusters)


def get_nc_cluster_scores(
    cluster_scores: dict[str, tuple[float, float]],
    cluster_sample: dict[str, str],
    min_comp: float,
    max_cont: float,
) -> dict[str, tuple[float, float]]:
    nc_cluster_scores: dict[str, tuple[float, float]] = dict()
    for cluster, scores in cluster_scores.items():
        comp, cont = scores
        comp, cont = float(comp), float(cont)
        comp, cont = comp / 100, cont / 100
        if cluster not in cluster_sample.keys():
            continue
        if comp >= min_comp and cont <= max_cont:
            nc_cluster_scores[cluster] = (comp, cont)

    return nc_cluster_scores


def get_cluster_sample(
    cluster_contigs: dict[str, set[str]], bin_separator: Optional[str]
) -> dict[str, str]:
    cluster_sample: dict[str, str] = dict()
    for cluster_ in cluster_contigs.keys():
        contigs = cluster_contigs[cluster_]
        contig_i = next(iter(contigs))
        sample = contig_i.split(bin_separator)[0]
        cluster_sample[cluster_] = sample

    return cluster_sample


def create_nc_sample_folders(
    cluster_scores: dict[str, tuple[float, float]],
    cluster_sample: dict[str, str],
    path_nc_bins_folder: str,
):
    nc_samples: set[str] = set()
    for cluster in cluster_scores.keys():
        sample = cluster_sample[cluster]
        nc_samples.add(sample)

    for sample in nc_samples:
        try:
            os.mkdir(os.path.join(path_nc_bins_folder, sample))
        except FileExistsError:
            pass


def write_nc_bins_from_mdrep_clusters(
    cluster_scores: dict[str, tuple[float, float]],
    cluster_sample: dict[str, str],
    path_nc_bins_folder: str,
    path_bins_folder: str,
):
    for cluster in cluster_scores.keys():
        sample = cluster_sample[cluster]
        src_bin = os.path.join(path_bins_folder, sample, cluster + ".fna")
        trg_bin = os.path.join(path_nc_bins_folder, sample, cluster + ".fna")
        shutil.move(src_bin, trg_bin)


def write_quality_report(
    cluster_scores: dict[str, tuple[float, float]], path_nc_bins_folder: str
):
    with open(os.path.join(path_nc_bins_folder, "quality_report.tsv"), "w") as file:
        print("Name completeness contamination", sep="\t", file=file)
        file.flush()
        for nc_cluster, (completeness, contaminaton) in cluster_scores.items():
            print(nc_cluster, completeness, contaminaton, sep="\t", file=file)
            file.flush()


def write_final_nc_clusters(
    cluster_scores: dict[str, tuple[float, float]],
    cluster_contigs: dict[str, set[str]],
    path_nc_clusters: str,
):
    with open(path_nc_clusters, "w") as file:
        print(vamb.vambtools.CLUSTERS_HEADER, file=file)
        for nc_cluster in cluster_scores.keys():
            nc_contigs = cluster_contigs[nc_cluster]
            for nc_contig in nc_contigs:
                print(nc_cluster, nc_contig, sep="\t", file=file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--c", type=str, help="path clusters file from tmp folder")
    parser.add_argument("--cf", type=str, help="path clusters file final")
    parser.add_argument("--cs_d", type=str, help="cluster_scores dictionary path ")
    parser.add_argument("--b", type=str, help="path all bins ")
    parser.add_argument(
        "--d", type=str, help="path to folder that will contain all nc bins"
    )
    parser.add_argument("--bin_separator", type=str, help="separator ")
    parser.add_argument("--comp", type=float, default=0.9, help="Min completeness ")
    parser.add_argument("--cont", type=float, default=0.05, help="Max contamination ")

    opt = parser.parse_args()

    with open(opt.c) as clusters_file:
        cluster_contigs = vamb.vambtools.read_clusters(clusters_file)

    with open(opt.cs_d) as f:
        cluster_scores = json.load(f)

    main(
        cluster_scores,
        cluster_contigs,
        opt.bin_separator,
        opt.d,
        opt.b,
        opt.cf,
        opt.comp,
        opt.cont,
    )



import unittest
import numpy as np
import random
import vamb


class TestAAE(unittest.TestCase):
    tnfs = np.random.random((111, 103)).astype(np.float32)
    rpkm = np.random.random((111, 14)).astype(np.float32)
    lens = np.random.randint(2000, 5000, size=111)
    contignames = ["".join(random.choices("abcdefghijklmnopqrstu", k=10)) for _ in lens]
    nlatent_l = 32
    default_args = (14, 256, nlatent_l, 25, 0.5, 0.5, 0.15, False, 0)
    default_temperature = 0.16

    # Construction
    def test_bad_args(self):
        default_args = self.default_args

        # Test the default args work
        aae = vamb.aamb_encode.AAE(*default_args)
        self.assertIsInstance(aae, vamb.aamb_encode.AAE)

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(0, *default_args[1:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:1], 0, *default_args[2:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:2], 0, *default_args[3:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:3], 0, *default_args[4:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:5], float("nan"), *default_args[6:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:5], -0.0001, *default_args[6:])

        with self.assertRaises(ValueError):
            vamb.aamb_encode.AAE(*default_args[:6], float("nan"), *default_args[7:])

    def test_loss_falls(self):
        aae = vamb.aamb_encode.AAE(*self.default_args)
        rpkm_copy = self.rpkm.copy()
        tnfs_copy = self.tnfs.copy()
        dl = vamb.encode.make_dataloader(
            rpkm_copy, tnfs_copy, self.lens, batchsize=16, destroy=True
        )
        (di, ti, ai, we) = next(iter(dl))
        mu, do, to, _, _, _, _ = aae(di, ti)
        start_loss = aae.calc_loss(di, do, ti, to)[0].data.item()

        # Loss drops with training
        aae.trainmodel(
            dl,
            nepochs=3,
            batchsteps=[1, 2],
            T=self.default_temperature,
            modelfile=None,
        )
        mu, do, to, _, _, _, _ = aae(di, ti)
        end_loss = aae.calc_loss(di, do, ti, to)[0].data.item()
        self.assertLess(end_loss, start_loss)

    def test_encode(self):
        aae = vamb.aamb_encode.AAE(*self.default_args)
        dl = vamb.encode.make_dataloader(
            self.rpkm.copy(), self.tnfs.copy(), self.lens, batchsize=16, destroy=True
        )
        (_, encoding) = aae.get_latents(self.contignames, dl)
        self.assertIsInstance(encoding, np.ndarray)
        self.assertEqual(encoding.dtype, np.float32)
        self.assertEqual(encoding.shape, (len(self.rpkm), self.nlatent_l))



#!/usr/bin/env python

import os
import argparse
import gzip
import vamb

parser = argparse.ArgumentParser(
    description="""Creates the input FASTA file for Vamb.
Input should be one or more FASTA files, each from a sample-specific assembly.
If keepnames is False, resulting FASTA can be binsplit with separator 'C'.""",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    add_help=True,
)

parser.add_argument("outpath", help="Path to output FASTA file")
parser.add_argument("inpaths", help="Paths to input FASTA file(s)", nargs="+")
parser.add_argument(
    "-m",
    dest="minlength",
    metavar="",
    type=int,
    default=2000,
    help="Discard sequences below this length [2000]",
)
parser.add_argument(
    "--keepnames", action="store_true", help="Do not rename sequences [False]"
)
parser.add_argument("--nozip", action="store_true", help="Do not gzip output [False]")

args = parser.parse_args()

# Check inputs
for path in args.inpaths:
    if not os.path.isfile(path):
        raise FileNotFoundError(path)

if os.path.exists(args.outpath):
    raise FileExistsError(args.outpath)

outpath = os.path.normpath(args.outpath)
parent = os.path.dirname(outpath)
if parent != "" and not os.path.isdir(parent):
    raise NotADirectoryError(
        f'Output file cannot be created: Parent directory "{parent}" is not an existing directory'
    )

# Run the code. Compressing DNA is easy, this is not much bigger than level 9, but
# many times faster
filehandle = (
    open(outpath, "w") if args.nozip else gzip.open(outpath, "wt", compresslevel=1)
)
try:
    vamb.vambtools.concatenate_fasta(
        filehandle, args.inpaths, minlength=args.minlength, rename=(not args.keepnames)
    )
except:
    filehandle.close()
    raise



#!/usr/bin/env python3

# More imports below, but the user's choice of processors must be parsed before
# numpy can be imported.
import vamb
import numpy as np
import sys
import os
import argparse
import torch
import time
import pycoverm
import itertools
from math import isfinite
from typing import Optional, Tuple, Union, cast, Callable, Literal, NamedTuple
from pathlib import Path
from collections.abc import Sequence
from torch.utils.data import DataLoader
from functools import partial
from loguru import logger

_ncpu = os.cpu_count()
DEFAULT_THREADS = 8 if _ncpu is None else min(_ncpu, 8)

# This is the smallest number of sequences that can be accepted.
# A smaller number than this cause issues either during normalization of the Abundance
# object, or causes PyTorch to throw dimension errors, as singleton dimensions
# are treated differently.
MINIMUM_SEQS = 2

# These MUST be set before importing numpy
# I know this is a shitty hack, see https://github.com/numpy/numpy/issues/11826
os.environ["MKL_NUM_THREADS"] = str(DEFAULT_THREADS)
os.environ["NUMEXPR_NUM_THREADS"] = str(DEFAULT_THREADS)
os.environ["OMP_NUM_THREADS"] = str(DEFAULT_THREADS)

# Append vamb to sys.path to allow vamb import even if vamb was not installed
# using pip
parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(parentdir)


def check_existing_file(path: Path) -> Path:
    if not path.is_file():
        raise FileNotFoundError(path)
    return path


def format_log(record) -> str:
    colors = {"WARNING": "red", "INFO": "green", "DEBUG": "blue", "ERROR": "red"}
    L = colors.get(record["level"].name, "blue")
    T = "red" if record["level"].name in ("WARNING", "ERROR") else "cyan"
    message = "<red>{message}</red>" if T == "red" else "{message}"
    time = f"<{T}>{{time:YYYY-MM-DD HH:mm:ss.SSS}}</{T}>"
    level = f"<b><{L}>{{level:<7}}</{L}></b>"
    return f"{time} | {level} | {message}\n{{exception}}"


def typeasserted(x, Ty: Union[type, tuple[type, ...]]):
    assert isinstance(x, Ty)
    return x


def try_make_dir(name: Union[Path, str]):
    try:
        os.mkdir(name)
    except FileExistsError:
        pass


class FASTAPath:
    def __init__(self, path: Path):
        self.path = check_existing_file(path)


class CompositionPath:
    def __init__(self, path: Path):
        self.path = check_existing_file(path)


class CompositionOptions:
    __slots__ = ["path"]

    @staticmethod
    def are_args_present(args: argparse.Namespace) -> bool:
        return isinstance(args.fasta, Path) or isinstance(args.composition, Path)

    @classmethod
    def from_args(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.fasta, (Path, type(None))),
            typeasserted(args.composition, (Path, type(None))),
        )

    def __init__(
        self,
        fastapath: Optional[Path],
        npzpath: Optional[Path],
    ):
        if not (fastapath is None) ^ (npzpath is None):
            raise argparse.ArgumentTypeError(
                "Must specify either FASTA or composition path"
            )

        for path in (fastapath, npzpath):
            if path is not None and not path.is_file():
                raise FileNotFoundError(path)

        if fastapath is not None:
            self.path = FASTAPath(fastapath)
        else:
            assert npzpath is not None
            self.path = CompositionPath(npzpath)


class AbundancePath:
    def __init__(self, path: Path):
        self.path = check_existing_file(path)


class BAMPaths:
    @classmethod
    def from_dir(cls, dir: Path, min_alignment_id: Optional[float]):
        if not dir.is_dir():
            raise NotADirectoryError(dir)
        paths = [b for b in dir.iterdir() if b.is_file() and b.suffix == ".bam"]
        return cls(paths, min_alignment_id)

    def __init__(self, paths: list[Path], min_alignment_id: Optional[float]):
        if len(paths) == 0:
            raise ValueError(
                "No BAM files passed. If `--bamdir` was used, check directory is not empty, "
                "and all BAM files ends with the extension `.bam`"
            )
        for path in paths:
            check_existing_file(path)
            if not pycoverm.is_bam_sorted(str(path)):
                raise ValueError(f"Path {path} is not sorted by reference.")

        if min_alignment_id is not None:
            if (
                not isfinite(min_alignment_id)
                or min_alignment_id < 0.0
                or min_alignment_id > 1.0
            ):
                raise argparse.ArgumentTypeError(
                    "Minimum nucleotide ID must be in [0,1]"
                )
            self.min_alignment_id = min_alignment_id
        else:
            self.min_alignment_id = 0.0
        self.paths = paths


class AbundanceTSVPath:
    __slots__ = ["path"]

    def __init__(self, path: Path):
        if not path.is_file():
            raise FileNotFoundError(path)
        self.path = path


class AbundanceOptions:
    __slots__ = ["paths"]

    @staticmethod
    def are_args_present(args: argparse.Namespace) -> bool:
        return (
            isinstance(args.bampaths, list)
            or isinstance(args.bamdir, Path)
            or isinstance(args.abundance_tsv, Path)
            or isinstance(args.abundancepath, Path)
        )

    @classmethod
    def from_args(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.bampaths, (list, type(None))),
            typeasserted(args.bamdir, (Path, type(None))),
            typeasserted(args.abundance_tsv, (Path, type(None))),
            typeasserted(args.abundancepath, (Path, type(None))),
            typeasserted(args.min_alignment_id, (float, type(None))),
        )

    def __init__(
        self,
        bampaths: Optional[list[Path]],
        bamdir: Optional[Path],
        abundance_tsv: Optional[Path],
        abundancepath: Optional[Path],
        min_alignment_id: Optional[float],
    ):
        # Make sure only one abundance input is there
        if (
            (bampaths is not None)
            + (abundancepath is not None)
            + (bamdir is not None)
            + (abundance_tsv is not None)
        ) != 1:
            raise argparse.ArgumentTypeError(
                "Must specify exactly one of BAM files, BAM dir, TSV file or abundance NPZ file input"
            )

        if abundancepath is not None:
            self.paths = AbundancePath(abundancepath)
        elif bamdir is not None:
            self.paths = BAMPaths.from_dir(bamdir, min_alignment_id)
        elif bampaths is not None:
            logger.warning(
                "The --bamfiles argument is deprecated. It works, but might be removed in future versions of Vamb. Please use --bamdir instead"
            )
            self.paths = BAMPaths(bampaths, min_alignment_id)
        elif abundance_tsv is not None:
            self.paths = AbundanceTSVPath(abundance_tsv)


class BasicTrainingOptions:
    __slots__ = ["num_epochs", "starting_batch_size", "batch_steps"]

    @classmethod
    def from_args_vae(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.nepochs, int),
            typeasserted(args.batchsize, int),
            typeasserted(args.batchsteps, list),
        )

    @classmethod
    def from_args_aae(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.nepochs_aae, int),
            typeasserted(args.batchsize_aae, int),
            typeasserted(args.batchsteps_aae, list),
        )

    @classmethod
    def from_args_taxometer(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.pred_nepochs, int),
            typeasserted(args.pred_batchsize, int),
            [],
        )

    def __init__(
        self, num_epochs: int, starting_batch_size: int, batch_steps: list[int]
    ):
        if num_epochs < 1:
            raise argparse.ArgumentTypeError(f"Minimum 1 epoch, not {num_epochs}")
        self.num_epochs = num_epochs

        if starting_batch_size < 1:
            raise argparse.ArgumentTypeError(
                f"Minimum batchsize of 1, not {starting_batch_size}"
            )
        self.starting_batch_size = starting_batch_size

        batch_steps = sorted(set(batch_steps))
        if max(batch_steps, default=0) >= self.num_epochs:
            raise argparse.ArgumentTypeError("All batchsteps must be less than nepochs")

        if min(batch_steps, default=1) < 1:
            raise argparse.ArgumentTypeError("All batchsteps must be 1 or higher")
        self.batch_steps = batch_steps


class VAEOptions:
    __slots__ = ["nhiddens", "nlatent", "alpha", "beta", "dropout", "basic_options"]

    @classmethod
    def from_args(cls, basic: BasicTrainingOptions, args: argparse.Namespace):
        return cls(
            basic,
            typeasserted(args.nhiddens, (list, type(None))),
            typeasserted(args.nlatent, int),
            typeasserted(args.alpha, (float, type(None))),
            typeasserted(args.beta, float),
            typeasserted(args.dropout, (float, type(None))),
        )

    def __init__(
        self,
        basic_options: BasicTrainingOptions,
        nhiddens: Optional[list[int]],
        nlatent: int,
        alpha: Optional[float],
        beta: float,
        dropout: Optional[float],
    ):
        self.basic_options = basic_options

        if nhiddens is not None and (
            len(nhiddens) == 0 or any(i < 1 for i in nhiddens)
        ):
            raise argparse.ArgumentTypeError(
                f"Minimum 1 neuron per layer, not {min(nhiddens)}"
            )
        self.nhiddens = nhiddens

        if nlatent < 1:
            raise argparse.ArgumentTypeError(f"Minimum 1 latent neuron, not {nlatent}")
        self.nlatent = nlatent

        if alpha is not None and ((not isfinite(alpha)) or alpha <= 0 or alpha >= 1):
            raise argparse.ArgumentTypeError("alpha must be above 0 and below 1")
        self.alpha = alpha

        if beta <= 0 or not isfinite(beta):
            raise argparse.ArgumentTypeError("beta cannot be negative or zero")
        self.beta = beta

        if dropout is not None and (dropout < 0 or dropout >= 1):
            raise argparse.ArgumentTypeError("dropout must be in 0 <= d < 1")
        self.dropout = dropout


class GeneralOptions:
    __slots__ = [
        "out_dir",
        "min_contig_length",
        "n_threads",
        "refcheck",
        "seed",
        "cuda",
    ]

    @classmethod
    def from_args(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.outdir, Path),
            typeasserted(args.minlength, int),
            typeasserted(args.nthreads, int),
            not typeasserted(args.norefcheck, bool),
            typeasserted(args.seed, int),
            typeasserted(args.cuda, bool),
        )

    def __init__(
        self,
        out_dir: Path,
        min_contig_length: int,
        n_threads: int,
        refcheck: bool,
        seed: int,
        cuda: bool,
    ):
        # Outdir does not exist
        if out_dir.exists():
            raise FileExistsError(out_dir)

        # Outdir is in an existing parent dir
        parent_dir = out_dir.parent
        if not parent_dir.is_dir():
            raise NotADirectoryError(parent_dir)
        self.out_dir = out_dir

        if n_threads < 1:
            raise ValueError(f"Must pass at least 1 thread, not {n_threads}")
        self.n_threads = n_threads

        if min_contig_length < 250:
            raise argparse.ArgumentTypeError(
                "Minimum contig length must be at least 250"
            )
        self.min_contig_length = min_contig_length

        if cuda and not torch.cuda.is_available():
            raise ModuleNotFoundError(
                "Cuda is not available on your PyTorch installation"
            )
        self.seed = seed
        self.cuda = cuda
        self.refcheck = refcheck


class TaxonomyBase:
    __slots__ = ["path"]

    def __init__(self, path: Path):
        self.path = path


class RefinedTaxonomy(TaxonomyBase):
    pass


class UnrefinedTaxonomy(TaxonomyBase):
    pass


def get_taxonomy(args: argparse.Namespace) -> Union[RefinedTaxonomy, UnrefinedTaxonomy]:
    path = args.taxonomy
    if path is None:
        raise ValueError(
            "Cannot load taxonomy for Taxometer without specifying --taxonomy"
        )
    with open(check_existing_file(path)) as file:
        try:
            header = next(file).rstrip("\r\n")
        except StopIteration:
            header = None

    if header is None:
        raise ValueError(f'Empty taxonomy path at "{path}"')
    elif header == vamb.taxonomy.TAXONOMY_HEADER:
        return UnrefinedTaxonomy(path)
    elif header == vamb.taxonomy.PREDICTED_TAXONOMY_HEADER:
        return RefinedTaxonomy(path)
    else:
        raise ValueError(
            f'ERROR: When reading taxonomy file at "{path}", '
            f"the first line was not either {repr(vamb.taxonomy.TAXONOMY_HEADER)} "
            f"or {repr(vamb.taxonomy.PREDICTED_TAXONOMY_HEADER)}'"
        )


class TaxometerOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        tax = get_taxonomy(args)
        if isinstance(tax, RefinedTaxonomy):
            raise ValueError(
                f'Attempted to run Taxometer to refine taxonomy at "{args.taxonomy}", '
                "but this file appears to already be an output of Taxometer"
            )
        return cls(
            GeneralOptions.from_args(args),
            CompositionOptions.from_args(args),
            AbundanceOptions.from_args(args),
            tax,
            BasicTrainingOptions.from_args_taxometer(args),
            typeasserted(args.pred_softmax_threshold, float),
            args.ploss,
        )

    def __init__(
        self,
        general: GeneralOptions,
        composition: CompositionOptions,
        abundance: AbundanceOptions,
        taxonomy: UnrefinedTaxonomy,
        basic: BasicTrainingOptions,
        softmax_threshold: float,
        ploss: Union[
            Literal["float_softmax"], Literal["cond_softmax"], Literal["soft_margin"]
        ],
    ):
        losses = ["flat_softmax", "cond_softmax", "soft_margin"]
        if (softmax_threshold > 1) or (softmax_threshold < 0):
            raise argparse.ArgumentTypeError(
                f"Softmax threshold should be between 0 and 1, currently {softmax_threshold}"
            )
        if ploss not in losses:
            raise argparse.ArgumentTypeError(
                f"Predictor loss needs to be one of {losses}"
            )
        self.general = general
        self.taxonomy = taxonomy
        self.ploss = ploss
        self.softmax_threshold = softmax_threshold
        self.composition = composition
        self.abundance = abundance
        self.basic = basic


class MarkerPath:
    def __init__(self, path: Path):
        self.path = check_existing_file(path)


class PredictMarkers:
    def __init__(self, hmm: Path, comp: CompositionOptions):
        self.hmm = check_existing_file(hmm)
        if not isinstance(comp.path, FASTAPath):
            raise ValueError(
                "If markers is predicted via --hmm, composition must be passed with --fasta"
            )
        self.fasta = comp.path


class MarkerOptions:
    @classmethod
    def from_args(cls, comp: CompositionOptions, args):
        markers = typeasserted(args.markers, (Path, type(None)))
        hmm = typeasserted(args.hmm_path, (Path, type(None)))
        if not (markers is None) ^ (hmm is None):
            raise ValueError(
                "Exactly one of --markers or --hmm_path must be set to input taxonomic information"
            )
        if markers is not None:
            return cls(MarkerPath(markers))
        else:
            assert isinstance(hmm, Path)  # we just checked this above
            return cls(PredictMarkers(hmm, comp))

    def __init__(self, markers: Union[MarkerPath, PredictMarkers]):
        self.content = markers


class KmeansOptions:
    def __init__(self, clusters: Path):
        self.clusters = check_existing_file(clusters)


class DBScanOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace, n_threads: int):
        tax = get_taxonomy(args)
        predict = not typeasserted(args.no_predictor, bool)
        if predict:
            if isinstance(tax, RefinedTaxonomy):
                logger.warning(
                    "Flag --no_predictor not set, but the taxonomy passed in "
                    "on --taxonomy is already refined. Skipped refinement."
                )
                return cls(tax, n_threads)
            else:
                if not (
                    CompositionOptions.are_args_present(args)
                    and AbundanceOptions.are_args_present(args)
                ):
                    raise ValueError(
                        "Flag --no_predictor is not set, but abundance "
                        "or composition has not been passed in, so there is no information "
                        "to refine taxonomy with"
                    )
                tax_options = TaxometerOptions(
                    GeneralOptions.from_args(args),
                    CompositionOptions.from_args(args),
                    AbundanceOptions.from_args(args),
                    tax,
                    BasicTrainingOptions.from_args_taxometer(args),
                    typeasserted(args.pred_softmax_threshold, float),
                    args.ploss,
                )
                return cls(tax_options, n_threads)
        else:
            return cls(tax, n_threads)

    def __init__(
        self,
        ops: Union[
            UnrefinedTaxonomy,
            RefinedTaxonomy,
            TaxometerOptions,
        ],
        n_processes: int,
    ):
        self.taxonomy = ops
        self.n_processes = n_processes


class ClusterOptions:
    __slots__ = [
        "window_size",
        "min_successes",
        "max_clusters",
    ]

    @classmethod
    def from_args(cls, args: argparse.Namespace):
        return cls(
            typeasserted(args.window_size, int),
            typeasserted(args.min_successes, int),
            typeasserted(args.max_clusters, (int, type(None))),
        )

    def __init__(
        self,
        window_size: int,
        min_successes: int,
        max_clusters: Optional[int],
    ):
        if window_size < 1:
            raise argparse.ArgumentTypeError("Window size must be at least 1")
        self.window_size = window_size

        if min_successes < 1 or min_successes > window_size:
            raise argparse.ArgumentTypeError(
                "Minimum cluster size must be in 1:windowsize"
            )
        self.min_successes = min_successes

        if max_clusters is not None and max_clusters < 1:
            raise argparse.ArgumentTypeError("Max clusters must be at least 1")
        self.max_clusters = max_clusters


class AAEOptions:
    @classmethod
    def from_args(
        cls,
        basic: BasicTrainingOptions,
        clustering: ClusterOptions,
        args: argparse.Namespace,
    ):
        return cls(
            basic,
            clustering,
            typeasserted(args.nhiddens_aae, int),
            typeasserted(args.nlatent_aae_z, int),
            typeasserted(args.nlatent_aae_y, int),
            typeasserted(args.sl, float),
            typeasserted(args.slr, float),
            typeasserted(args.temp, float),
        )

    def __init__(
        self,
        basic: BasicTrainingOptions,
        clustering: ClusterOptions,
        nhiddens: int,
        nlatent_z: int,
        nlatent_y: int,
        sl: float,
        slr: float,
        temp: float,
    ):
        # For the VAE-AAE workflow, we must cluster the full set of sequences,
        # else the cluster dereplication step makes no sense, as the different
        # clusterings will contain different sets of clusters.
        # So, enforce this here
        if clustering.max_clusters is not None:
            raise ValueError(
                "When using the VAE-AAE model, `max_clusters` (option `-c`) "
                "must not be set."
            )

        self.basic = basic
        if nhiddens < 1:
            raise argparse.ArgumentTypeError(
                f"Minimum 1 neuron in the hidden layers, not {nhiddens}"
            )

        self.nhiddens = nhiddens

        if nlatent_z < 1:
            raise argparse.ArgumentTypeError(
                f"Minimum 1 latent neuron, not {nlatent_z}"
            )
        self.nlatent_z = nlatent_z

        if nlatent_y < 1:
            raise argparse.ArgumentTypeError(
                f"Minimum 1 latent neuron, not {nlatent_y}"
            )
        self.nlatent_y = nlatent_y
        self.temp = temp
        self.sl = sl
        self.slr = slr


class BinOutputOptions:
    __slots__ = ["binsplitter", "min_fasta_output_size"]

    # We take a composition as arguments, because if min_fasta_output_size is set,
    # we need to guarantee that the composition is passed as fasta.
    # Similarly, we currently don't have any workflows that can use BinOutputOptions
    # without using composition.

    @classmethod
    def from_args(cls, comp: CompositionOptions, args: argparse.Namespace):
        return cls(
            comp,
            typeasserted(args.binsplit_separator, (str, type(None))),
            typeasserted(args.min_fasta_output_size, (int, type(None))),
        )

    def __init__(
        self,
        composition: CompositionOptions,
        binsplit_separator: Optional[str],
        min_fasta_output_size: Optional[int],
    ):
        self.binsplitter = vamb.vambtools.BinSplitter(binsplit_separator)
        if min_fasta_output_size is not None:
            if not isinstance(composition.path, FASTAPath):
                raise argparse.ArgumentTypeError(
                    "If minfasta is not None, input fasta file must be given explicitly"
                )
            if min_fasta_output_size < 0:
                raise argparse.ArgumentTypeError(
                    "Minimum FASTA output size must be nonnegative"
                )
        self.min_fasta_output_size = min_fasta_output_size


@logger.catch(reraise=True)
def run(
    runner: Callable[[], None],
    general: GeneralOptions,
):
    torch.set_num_threads(general.n_threads)
    try_make_dir(general.out_dir)
    logger.add(general.out_dir.joinpath("log.txt"), format=format_log)
    begintime = time.time()
    logger.info("Starting Vamb version " + vamb.__version_str__)
    logger.info("Random seed is " + str(general.seed))
    logger.info(f"Invoked with CLI args: '{' '.join(sys.argv)}'")
    runner()
    logger.info(f"Completed Vamb in {round(time.time() - begintime, 2)} seconds.")


class BinnerCommonOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        comp = CompositionOptions.from_args(args)
        return cls(
            GeneralOptions.from_args(args),
            comp,
            AbundanceOptions.from_args(args),
            ClusterOptions.from_args(args),
            BinOutputOptions.from_args(comp, args),
        )

    # We do not have BasicTrainingOptions because that is model-specific
    def __init__(
        self,
        general: GeneralOptions,
        comp: CompositionOptions,
        abundance: AbundanceOptions,
        clustering: ClusterOptions,
        output: BinOutputOptions,
    ):
        self.general = general
        self.comp = comp
        self.abundance = abundance
        self.clustering = clustering
        self.output = output


class BinDefaultOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        common = BinnerCommonOptions.from_args(args)
        basic = BasicTrainingOptions.from_args_vae(args)
        return cls(
            common,
            VAEOptions.from_args(basic, args),
        )

    def __init__(
        self,
        common: BinnerCommonOptions,
        vae: VAEOptions,
    ):
        self.common = common
        self.vae = vae


class BinTaxVambOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        common = BinnerCommonOptions.from_args(args)
        basic = BasicTrainingOptions.from_args_vae(args)
        vae = VAEOptions.from_args(basic, args)
        taxonomy = get_taxonomy(args)
        predict = not typeasserted(args.no_predictor, bool)
        if predict:
            if isinstance(taxonomy, RefinedTaxonomy):
                logger.warning(
                    "Flag --no_predictor not set, but the taxonomy passed in "
                    "on --taxonomy is already refined. Skipped refinement."
                )
                tax = taxonomy
            else:
                tax = TaxometerOptions(
                    common.general,
                    common.comp,
                    common.abundance,
                    taxonomy,
                    basic,
                    typeasserted(args.pred_softmax_threshold, float),
                    args.ploss,
                )
        else:
            tax = taxonomy
        return cls(common, vae, tax)

    # The VAEVAE models share the same settings as the VAE model, so we just use VAEOptions
    def __init__(
        self,
        common: BinnerCommonOptions,
        vae: VAEOptions,
        taxonomy: Union[RefinedTaxonomy, TaxometerOptions, UnrefinedTaxonomy],
    ):
        self.common = common
        self.vae = vae
        self.taxonomy = taxonomy


class BinAvambOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        common = BinnerCommonOptions.from_args(args)
        basic_vae = BasicTrainingOptions.from_args_vae(args)
        basic_aae = BasicTrainingOptions.from_args_aae(args)
        clustering = ClusterOptions.from_args(args)
        return cls(
            common,
            VAEOptions.from_args(basic_vae, args),
            AAEOptions.from_args(basic_aae, clustering, args),
        )

    def __init__(
        self,
        common: BinnerCommonOptions,
        vae: VAEOptions,
        aae: AAEOptions,
    ):
        self.common = common
        self.vae = vae
        self.aae = aae


class ReclusteringOptions:
    @classmethod
    def from_args(cls, args: argparse.Namespace):
        general = GeneralOptions.from_args(args)
        latent = args.latent_path
        if latent is None:
            raise ValueError("--latent_path must be set")
        if args.algorithm == "kmeans":
            clusters = args.clusters_path
            if clusters is None:
                raise ValueError(
                    "If --algorithm is set to 'kmeans', --clusters_path must be set"
                )
            algorithm = KmeansOptions(clusters)
        elif args.algorithm == "dbscan":
            algorithm = DBScanOptions.from_args(args, general.n_threads)
        else:
            assert False  # no more algorithms

        # Avoid loading composition again if already loaded in DBScanOptions
        if isinstance(algorithm, DBScanOptions) and isinstance(
            algorithm.taxonomy, tuple
        ):
            comp = algorithm.taxonomy[1]
        else:
            comp = CompositionOptions.from_args(args)

        markers = MarkerOptions.from_args(comp, args)
        output = BinOutputOptions.from_args(comp, args)
        return cls(
            latent,
            comp,
            markers,
            general,
            output,
            algorithm,
        )

    def __init__(
        self,
        latent_path: Path,
        comp: CompositionOptions,
        markers: MarkerOptions,
        general: GeneralOptions,
        output: BinOutputOptions,
        algorithm: Union[KmeansOptions, DBScanOptions],
    ):
        self.latent_path = check_existing_file(latent_path)
        self.composition = comp
        self.markers = markers
        self.general = general
        self.output = output
        self.algorithm = algorithm


def calc_tnf(
    options: CompositionOptions,
    min_contig_length: int,
    outdir: Path,
    binsplitter: vamb.vambtools.BinSplitter,
) -> vamb.parsecontigs.Composition:
    begintime = time.time()
    logger.info("Loading TNF")
    logger.info(f"\tMinimum sequence length: {min_contig_length}")

    path = options.path

    if isinstance(path, CompositionPath):
        logger.info(f'\tLoading composition from npz at: "{path.path}"')
        composition = vamb.parsecontigs.Composition.load(path.path)
        composition.filter_min_length(min_contig_length)
    else:
        assert isinstance(path, FASTAPath)
        logger.info(f"\tLoading data from FASTA file {path.path}")
        with vamb.vambtools.Reader(path.path) as file:
            composition = vamb.parsecontigs.Composition.from_file(
                file, minlength=min_contig_length
            )
        composition.save(outdir.joinpath("composition.npz"))

    binsplitter.initialize(composition.metadata.identifiers)

    if composition.nseqs < MINIMUM_SEQS:
        err = (
            f"Found only {composition.nseqs} contigs, but Vamb currently requires at least "
            f"{MINIMUM_SEQS} to work correctly. "
            "If you have this few sequences in a metagenomic assembly, "
            "it's probably an error somewhere in your workflow."
        )
        logger.error(err)
        raise ValueError(err)

    # Warn the user if any contigs have been observed, which is smaller
    # than the threshold.
    if not np.all(composition.metadata.mask):
        n_removed = len(composition.metadata.mask) - np.sum(composition.metadata.mask)
        message = (
            f"The minimum sequence length has been set to {min_contig_length}, "
            f"but {n_removed} sequences fell below this threshold and was filtered away."
            "\nBetter results are obtained if the sequence file is filtered to the minimum "
            "sequence length before mapping.\n"
        )
        logger.opt(raw=True).info("\n")
        logger.warning(message)

    elapsed = round(time.time() - begintime, 2)
    logger.info(
        f"\tKept {composition.count_bases()} bases in {composition.nseqs} sequences"
    )
    logger.info(f"\tProcessed TNF in {elapsed} seconds.\n")

    return composition


def calc_abundance(
    abundance_options: AbundanceOptions,
    outdir: Path,
    refcheck: bool,
    comp_metadata: vamb.parsecontigs.CompositionMetaData,
    nthreads: int,
) -> vamb.parsebam.Abundance:
    begintime = time.time()
    logger.info("Loading depths")
    logger.info(
        f"\tReference hash: {comp_metadata.refhash.hex() if refcheck else 'None'}"
    )

    paths = abundance_options.paths
    if isinstance(paths, AbundancePath):
        logger.info(f'\tLoading depths from npz at: "{str(paths.path)}"')

        abundance = vamb.parsebam.Abundance.load(
            paths.path,
            comp_metadata.refhash if refcheck else None,
        )
        # I don't want this check in any constructors of abundance, since the constructors
        # should be able to skip this check in case comp and abundance are independent.
        # But when running the main Vamb workflow, we need to assert this.
        if abundance.nseqs != comp_metadata.nseqs:
            assert not refcheck
            raise ValueError(
                f"Loaded abundance has {abundance.nseqs} sequences, "
                f"but composition has {comp_metadata.nseqs}."
            )
    elif isinstance(paths, BAMPaths):
        logger.info(f"\tParsing {len(paths.paths)} BAM files with {nthreads} threads")
        logger.info(f"\tMin identity: {paths.min_alignment_id}")

        abundance = vamb.parsebam.Abundance.from_files(
            list(paths.paths),
            outdir.joinpath("tmp").joinpath("pycoverm"),
            comp_metadata,
            refcheck,
            paths.min_alignment_id,
            nthreads,
        )
        abundance.save(outdir.joinpath("abundance.npz"))

        logger.info("\tOrder of columns is:")
        for i, samplename in enumerate(abundance.samplenames):
            logger.info(f"\t{i:>6}: {samplename}")
    elif isinstance(paths, AbundanceTSVPath):
        logger.info(f'\tParsing abundance from TSV at "{paths.path}"')
        abundance = vamb.parsebam.Abundance.from_tsv(
            paths.path,
            comp_metadata,
        )
        abundance.save(outdir.joinpath("abundance.npz"))
        logger.info("\tOrder of columns is:")
        for i, samplename in enumerate(abundance.samplenames):
            logger.info(f"\t{i:>6}: {samplename}")

    else:
        assert False

    elapsed = round(time.time() - begintime, 2)
    logger.info(f"\tProcessed abundance in {elapsed} seconds.\n")

    return abundance


def load_composition_and_abundance(
    vamb_options: GeneralOptions,
    comp_options: CompositionOptions,
    abundance_options: AbundanceOptions,
    binsplitter: vamb.vambtools.BinSplitter,
) -> Tuple[vamb.parsecontigs.Composition, vamb.parsebam.Abundance]:
    composition = calc_tnf(
        comp_options, vamb_options.min_contig_length, vamb_options.out_dir, binsplitter
    )
    abundance = calc_abundance(
        abundance_options,
        vamb_options.out_dir,
        vamb_options.refcheck,
        composition.metadata,
        vamb_options.n_threads,
    )
    return (composition, abundance)


def load_markers(
    options: MarkerOptions,
    comp_metadata: vamb.parsecontigs.CompositionMetaData,
    existing_outdir: Path,
    n_threads: int,
) -> vamb.parsemarkers.Markers:
    logger.info("Loading markers")
    begin_time = time.time()
    if isinstance(options.content, MarkerPath):
        path = options.content.path
        logger.info(f'\tLoading markers from existing `markers.npz` at "{path}"')
        markers = vamb.parsemarkers.Markers.load(path, comp_metadata.refhash)
    elif isinstance(options.content, PredictMarkers):
        logger.info("\tPredicting markers. This might take some time")
        logger.info(f"\t\tFASTA file located at {options.content.fasta.path}")
        logger.info(
            f"\t\tHMM profile file (.hmm file) located at {options.content.hmm}"
        )
        markers = vamb.parsemarkers.Markers.from_files(
            options.content.fasta.path,
            options.content.hmm,
            cast(Sequence[str], comp_metadata.identifiers),
            existing_outdir.joinpath("tmp_markers"),
            n_threads,
            comp_metadata.refhash,
        )
        markers.save(existing_outdir.joinpath("markers.npz"))
    else:
        assert False

    elapsed = round(time.time() - begin_time, 2)
    logger.info(f"\tProcessed markers in {elapsed} seconds.\n")
    return markers


def trainvae(
    vae_options: VAEOptions,
    vamb_options: GeneralOptions,
    data_loader: DataLoader,
) -> np.ndarray:
    begintime = time.time()
    logger.info("Creating and training VAE")

    nsamples = data_loader.dataset.tensors[0].shape[1]  # type:ignore
    vae = vamb.encode.VAE(
        nsamples,
        nhiddens=vae_options.nhiddens,
        nlatent=vae_options.nlatent,
        alpha=vae_options.alpha,
        beta=vae_options.beta,
        dropout=vae_options.dropout,
        cuda=vamb_options.cuda,
        seed=vamb_options.seed,
    )

    logger.info("\tCreated VAE")
    modelpath = vamb_options.out_dir.joinpath("model.pt")
    vae.trainmodel(
        vamb.encode.set_batchsize(
            data_loader, vae_options.basic_options.starting_batch_size
        ),
        nepochs=vae_options.basic_options.num_epochs,
        batchsteps=vae_options.basic_options.batch_steps,
        modelfile=modelpath,
    )

    logger.info("\tEncoding to latent representation")
    latent = vae.encode(data_loader)
    vamb.vambtools.write_npz(vamb_options.out_dir.joinpath("latent.npz"), latent)
    del vae  # Needed to free "latent" array's memory references?

    elapsed = round(time.time() - begintime, 2)
    logger.info(f"\tTrained VAE and encoded in {elapsed} seconds.\n")

    return latent


def trainaae(
    data_loader: DataLoader,
    aae_options: AAEOptions,
    vamb_options: GeneralOptions,
    alpha: Optional[float],  # set automatically if None
    contignames: Sequence[str],
) -> tuple[np.ndarray, dict[str, set[str]]]:
    begintime = time.time()
    logger.info("Creating and training AAE")
    nsamples = data_loader.dataset.tensors[0].shape[1]  # type:ignore

    aae = vamb.aamb_encode.AAE(
        nsamples,
        aae_options.nhiddens,
        aae_options.nlatent_z,
        aae_options.nlatent_y,
        aae_options.sl,
        aae_options.slr,
        alpha,
        _cuda=vamb_options.cuda,
        seed=vamb_options.seed,
    )

    logger.info("\tCreated AAE")
    modelpath = os.path.join(vamb_options.out_dir, "aae_model.pt")
    aae.trainmodel(
        vamb.encode.set_batchsize(data_loader, aae_options.basic.starting_batch_size),
        aae_options.basic.num_epochs,
        aae_options.basic.batch_steps,
        aae_options.temp,
        modelpath,
    )

    logger.info("\tEncoding to latent representation")
    clusters_y_dict, latent = aae.get_latents(contignames, data_loader)
    vamb.vambtools.write_npz(
        os.path.join(vamb_options.out_dir, "aae_z_latent.npz"), latent
    )

    del aae  # Needed to free "latent" array's memory references?

    elapsed = round(time.time() - begintime, 2)
    logger.info(f"\tTrained AAE and encoded in {elapsed} seconds.\n")

    return latent, clusters_y_dict


class FastaOutput(NamedTuple):
    existing_fasta_path: FASTAPath
    bins_dir_to_populate: Path  # (or to create, if not existing)
    min_fasta_size: int

    @classmethod
    def try_from_common(cls, common: BinnerCommonOptions):
        if common.output.min_fasta_output_size is not None:
            # We have checked this in the BinOutputOptions constructor
            assert isinstance(common.comp.path, FASTAPath)
            return cls(
                common.comp.path,
                common.general.out_dir.joinpath("bins"),
                common.output.min_fasta_output_size,
            )
        else:
            return None


def cluster_and_write_files(
    cluster_options: ClusterOptions,
    binsplitter: vamb.vambtools.BinSplitter,
    latent: np.ndarray,
    sequence_names: Sequence[str],
    sequence_lens: np.ndarray,
    seed: int,
    cuda: bool,
    base_clusters_name: str,  # e.g. /foo/bar/vae -> /foo/bar/vae_unsplit.tsv
    fasta_output: Optional[FastaOutput],
    bin_prefix: Optional[str],  # see write_clusters_and_bins
):
    begintime = time.time()
    # Create cluser iterator
    logger.info("Clustering")
    logger.info(f"\tWindowsize: {cluster_options.window_size}")
    logger.info(
        f"\tMin successful thresholds detected: {cluster_options.min_successes}",
    )
    logger.info(f"\tMax clusters: {cluster_options.max_clusters}")
    logger.info(f"\tUse CUDA for clustering: {cuda}")
    logger.info(f"\tBinsplitter: {binsplitter.log_string()}")

    cluster_generator = vamb.cluster.ClusterGenerator(
        latent,
        sequence_lens,
        windowsize=cluster_options.window_size,
        minsuccesses=cluster_options.min_successes,
        destroy=True,
        normalized=False,
        cuda=cuda,
        rng_seed=seed,
    )
    # This also works correctly when max_clusters is None
    clusters = itertools.islice(cluster_generator, cluster_options.max_clusters)
    cluster_dict: dict[str, set[str]] = dict()

    # Write the cluster metadata to file
    with open(Path(base_clusters_name + "_metadata.tsv"), "w") as file:
        print("name\tradius\tpeak valley ratio\tkind\tbp\tncontigs\tmedoid", file=file)
        for i, cluster in enumerate(clusters):
            cluster_dict[str(i + 1)] = {
                sequence_names[cast(int, i)] for i in cluster.members
            }
            print(
                str(i + 1),
                None if cluster.radius is None else round(cluster.radius, 3),
                (
                    None
                    if cluster.observed_pvr is None
                    else round(cluster.observed_pvr, 2)
                ),
                cluster.kind_str,
                sum(sequence_lens[i] for i in cluster.members),
                len(cluster.members),
                sequence_names[cluster.medoid],
                file=file,
                sep="\t",
            )

    elapsed = round(time.time() - begintime, 2)

    write_clusters_and_bins(
        fasta_output,
        bin_prefix,
        binsplitter,
        base_clusters_name,
        cluster_dict,
        sequence_names,
        cast(Sequence[int], sequence_lens),
    )
    logger.info(f"\tClustered contigs in {elapsed} seconds.\n")


def write_clusters_and_bins(
    fasta_output: Optional[FastaOutput],
    # If `x` and not None, clusters will be renamed `x` + old_name.
    # This is necessary since for the AAE, we may need to write bins
    # from three latent spaces into the same directory, and the names
    # must not clash.
    bin_prefix: Optional[str],
    binsplitter: vamb.vambtools.BinSplitter,
    base_clusters_name: str,  # e.g. /foo/bar/vae -> /foo/bar/vae_unsplit.tsv
    clusters: dict[str, set[str]],
    sequence_names: Sequence[str],
    sequence_lens: Sequence[int],
):
    # Write unsplit clusters to file
    begintime = time.time()
    unsplit_path = Path(base_clusters_name + "_unsplit.tsv")
    with open(unsplit_path, "w") as file:
        (n_unsplit_clusters, n_contigs) = vamb.vambtools.write_clusters(
            file, clusters.items()
        )

    # Open unsplit clusters and split them
    if binsplitter.splitter is not None:
        split_path = Path(base_clusters_name + "_split.tsv")
        clusters = dict(binsplitter.binsplit(clusters.items()))
        # Add prefix before writing the clusters to file
        clusters = add_bin_prefix(clusters, bin_prefix)
        with open(split_path, "w") as file:
            (n_split_clusters, _) = vamb.vambtools.write_clusters(
                file, clusters.items()
            )
        msg = f"\tClustered {n_contigs} contigs in {n_split_clusters} split bins ({n_unsplit_clusters} clusters)"
    else:
        msg = f"\tClustered {n_contigs} contigs in {n_unsplit_clusters} unsplit bins"
        clusters = add_bin_prefix(clusters, bin_prefix)

    logger.info(msg)
    elapsed = round(time.time() - begintime, 2)
    logger.info(f"\tWrote cluster file(s) in {elapsed} seconds.")

    # Write bins, if necessary
    if fasta_output is not None:
        starttime = time.time()
        filtered_clusters: dict[str, set[str]] = dict()
        assert len(sequence_lens) == len(sequence_names)
        sizeof = dict(zip(sequence_names, sequence_lens))
        for binname, contigs in clusters.items():
            if sum(sizeof[c] for c in contigs) >= fasta_output.min_fasta_size:
                filtered_clusters[binname] = contigs

        with vamb.vambtools.Reader(fasta_output.existing_fasta_path.path) as file:
            vamb.vambtools.write_bins(
                fasta_output.bins_dir_to_populate,
                filtered_clusters,
                file,
                None,
            )
        elapsed = round(time.time() - starttime, 2)
        n_bins = len(filtered_clusters)
        n_contigs = sum(len(v) for v in filtered_clusters.values())
        logger.info(
            f"\tWrote {n_bins} bins with {n_contigs} sequences in {elapsed} seconds."
        )


def add_bin_prefix(
    clusters: dict[str, set[str]], prefix: Optional[str]
) -> dict[str, set[str]]:
    if prefix is None:
        return clusters
    else:
        return {prefix + b: c for (b, c) in clusters.items()}


def run_bin_default(opt: BinDefaultOptions):
    composition, abundance = load_composition_and_abundance(
        vamb_options=opt.common.general,
        comp_options=opt.common.comp,
        abundance_options=opt.common.abundance,
        binsplitter=opt.common.output.binsplitter,
    )
    data_loader = vamb.encode.make_dataloader(
        abundance.matrix,
        composition.matrix,
        composition.metadata.lengths,
        256,  # dummy value - we change this before using the actual loader
        destroy=True,
        cuda=opt.common.general.cuda,
    )
    latent = trainvae(
        vae_options=opt.vae,
        vamb_options=opt.common.general,
        data_loader=data_loader,
    )
    # Free up memory
    comp_metadata = composition.metadata
    del composition, abundance
    assert comp_metadata.nseqs == len(latent)

    cluster_and_write_files(
        opt.common.clustering,
        opt.common.output.binsplitter,
        latent,
        cast(Sequence[str], comp_metadata.identifiers),
        comp_metadata.lengths,
        opt.common.general.seed,
        opt.common.general.cuda,
        str(opt.common.general.out_dir.joinpath("vae_clusters")),
        FastaOutput.try_from_common(opt.common),
        None,
    )
    del latent


def run_bin_aae(opt: BinAvambOptions):
    composition, abundance = load_composition_and_abundance(
        vamb_options=opt.common.general,
        comp_options=opt.common.comp,
        abundance_options=opt.common.abundance,
        binsplitter=opt.common.output.binsplitter,
    )
    data_loader = vamb.encode.make_dataloader(
        abundance.matrix,
        composition.matrix,
        composition.metadata.lengths,
        256,  # dummy value - we change this before using the actual loader
        destroy=True,
        cuda=opt.common.general.cuda,
    )
    latent_z, clusters_y_dict = trainaae(
        data_loader=data_loader,
        aae_options=opt.aae,
        vamb_options=opt.common.general,
        alpha=opt.vae.alpha,
        contignames=cast(Sequence[str], composition.metadata.identifiers),
    )
    # Free up memory
    comp_metadata = composition.metadata
    del composition, abundance
    assert comp_metadata.nseqs == len(latent_z)
    # Cluster and output the Z clusters
    # This function calls write_clusters_and_bins,
    # but also does the actual clustering and writes cluster metadata.
    # This does not apply to the aae_y clusters, since their cluster label
    # can be extracted directly from the latent space without clustering,
    # and hence below, `write_clusters_and_bins` is called directly instead.
    cluster_and_write_files(
        opt.common.clustering,
        opt.common.output.binsplitter,
        latent_z,
        cast(Sequence[str], comp_metadata.identifiers),
        comp_metadata.lengths,
        opt.common.general.seed,
        opt.common.general.cuda,
        str(opt.common.general.out_dir.joinpath("aae_z_clusters")),
        FastaOutput.try_from_common(opt.common),
        "z_",
    )
    del latent_z

    # We enforce this in the VAEAAEOptions constructor, see comment there
    # Cluster and output the Y clusters
    assert opt.common.clustering.max_clusters is None
    write_clusters_and_bins(
        FastaOutput.try_from_common(opt.common),
        "y_",
        binsplitter=opt.common.output.binsplitter,
        base_clusters_name=str(opt.common.general.out_dir.joinpath("aae_y_clusters")),
        clusters=clusters_y_dict,
        sequence_names=cast(Sequence[str], comp_metadata.identifiers),
        sequence_lens=cast(Sequence[int], comp_metadata.lengths),
    )


def predict_taxonomy(
    comp_metadata: vamb.parsecontigs.CompositionMetaData,
    abundance: np.ndarray,
    tnfs: np.ndarray,
    lengths: np.ndarray,
    out_dir: Path,
    taxonomy_options: TaxometerOptions,
    cuda: bool,
) -> vamb.taxonomy.PredictedTaxonomy:
    begintime = time.time()
    logger.info("Predicting taxonomy with Taxometer")

    taxonomies = vamb.taxonomy.Taxonomy.from_file(
        taxonomy_options.taxonomy.path, comp_metadata, False
    )
    nodes, ind_nodes, table_parent = vamb.taxvamb_encode.make_graph(
        taxonomies.contig_taxonomies
    )
    logger.info(f"\t{len(nodes)} nodes in the graph")
    classes_order: list[str] = []
    for i in taxonomies.contig_taxonomies:
        if i is None or len(i.ranks) == 0:
            classes_order.append("root")
        else:
            classes_order.append(i.ranks[-1])
    targets = np.array([ind_nodes[i] for i in classes_order])

    model = vamb.taxvamb_encode.VAMB2Label(
        abundance.shape[1],
        len(nodes),
        nodes,
        table_parent,
        nhiddens=[512, 512, 512, 512],
        hier_loss=taxonomy_options.ploss,
        cuda=cuda,
    )

    dataloader_vamb = vamb.encode.make_dataloader(
        abundance,
        tnfs,
        lengths,
        batchsize=taxonomy_options.basic.starting_batch_size,
    )
    dataloader_joint = vamb.taxvamb_encode.make_dataloader_concat_hloss(
        abundance,
        tnfs,
        lengths,
        targets,
        len(nodes),
        table_parent,
        batchsize=taxonomy_options.basic.starting_batch_size,
    )

    logger.info("\tCreated dataloader")

    predictortime = time.time()
    logger.info("Starting training the taxonomy predictor")
    logger.info(f"Using threshold {taxonomy_options.softmax_threshold}")

    model_path = out_dir.joinpath("predictor_model.pt")
    with open(model_path, "wb") as modelfile:
        model.trainmodel(
            dataloader_joint,
            nepochs=taxonomy_options.basic.num_epochs,
            modelfile=modelfile,
            batchsteps=taxonomy_options.basic.batch_steps,
        )
    logger.info(
        f"Finished training the taxonomy predictor in {round(time.time() - predictortime, 2)} seconds."
    )

    logger.info("Writing the taxonomy predictions")
    predicted_path = out_dir.joinpath("results_taxometer.tsv")
    nodes_ar = np.array(nodes)

    logger.info(f"Using threshold {taxonomy_options.softmax_threshold}")
    contig_taxonomies: list[vamb.taxonomy.PredictedContigTaxonomy] = []
    for predicted_vector, predicted_labels in model.predict(dataloader_vamb):
        for i in range(predicted_vector.shape[0]):
            label = predicted_labels[i]
            while table_parent[label] != -1:
                label = table_parent[label]
            threshold_mask: Sequence[bool] = (
                predicted_vector[i] > taxonomy_options.softmax_threshold
            )
            ranks = list(nodes_ar[threshold_mask][1:])
            probs = predicted_vector[i][threshold_mask][1:]
            tax = vamb.taxonomy.PredictedContigTaxonomy(
                vamb.taxonomy.ContigTaxonomy(ranks), probs
            )
            contig_taxonomies.append(tax)

    taxonomy = vamb.taxonomy.PredictedTaxonomy(contig_taxonomies, comp_metadata, False)

    with open(predicted_path, "w") as file:
        taxonomy.write_as_tsv(file, comp_metadata)

    logger.info(
        f"Completed taxonomy predictions in {round(time.time() - begintime, 2)} seconds."
    )
    return taxonomy


def run_taxonomy_predictor(opt: TaxometerOptions):
    composition, abundance = load_composition_and_abundance(
        opt.general,
        opt.composition,
        opt.abundance,
        vamb.vambtools.BinSplitter.inert_splitter(),
    )
    abundance, tnfs, lengths, _ = (
        abundance.matrix,
        composition.matrix,
        composition.metadata.lengths,
        composition.metadata.identifiers,
    )
    predict_taxonomy(
        comp_metadata=composition.metadata,
        abundance=abundance,
        tnfs=tnfs,
        lengths=lengths,
        out_dir=opt.general.out_dir,
        taxonomy_options=opt,
        cuda=opt.general.cuda,
    )


def run_vaevae(opt: BinTaxVambOptions):
    vae_options = opt.vae
    composition, abundance = load_composition_and_abundance(
        vamb_options=opt.common.general,
        comp_options=opt.common.comp,
        abundance_options=opt.common.abundance,
        binsplitter=opt.common.output.binsplitter,
    )
    abundance, tnfs, lengths, contignames = (
        abundance.matrix,
        composition.matrix,
        composition.metadata.lengths,
        composition.metadata.identifiers,
    )
    if isinstance(opt.taxonomy, TaxometerOptions):
        predicted_contig_taxonomies = predict_taxonomy(
            comp_metadata=composition.metadata,
            abundance=abundance,
            tnfs=tnfs,
            lengths=lengths,
            out_dir=opt.common.general.out_dir,
            taxonomy_options=opt.taxonomy,
            cuda=opt.common.general.cuda,
        )
        contig_taxonomies = predicted_contig_taxonomies.to_taxonomy()
    elif isinstance(opt.taxonomy, RefinedTaxonomy):
        logger.info(f'Loading already-refined taxonomy from file "{opt.taxonomy.path}"')
        contig_taxonomies = vamb.taxonomy.Taxonomy.from_refined_file(
            opt.taxonomy.path, composition.metadata, False
        )
    else:
        logger.info(f'Loading unrefined taxonomy from file "{opt.taxonomy.path}"')
        contig_taxonomies = vamb.taxonomy.Taxonomy.from_file(
            opt.taxonomy.path, composition.metadata, False
        )

    nodes, ind_nodes, table_parent = vamb.taxvamb_encode.make_graph(
        contig_taxonomies.contig_taxonomies
    )
    classes_order: list[str] = []
    for i in contig_taxonomies.contig_taxonomies:
        if i is None or len(i.ranks) == 0:
            classes_order.append("root")
        else:
            classes_order.append(i.ranks[-1])
    targets = np.array([ind_nodes[i] for i in classes_order])

    vae = vamb.taxvamb_encode.VAEVAEHLoss(
        abundance.shape[1],
        len(nodes),
        nodes,
        table_parent,
        nhiddens=vae_options.nhiddens,
        nlatent=vae_options.nlatent,
        alpha=vae_options.alpha,
        beta=vae_options.beta,
        dropout=vae_options.dropout,
        cuda=opt.common.general.cuda,
    )

    dataloader_vamb = vamb.encode.make_dataloader(
        abundance,
        tnfs,
        lengths,
        batchsize=vae_options.basic_options.starting_batch_size,
        cuda=opt.common.general.cuda,
    )
    dataloader_joint = vamb.taxvamb_encode.make_dataloader_concat_hloss(
        abundance,
        tnfs,
        lengths,
        targets,
        len(nodes),
        table_parent,
        batchsize=vae_options.basic_options.starting_batch_size,
        cuda=opt.common.general.cuda,
    )
    dataloader_labels = vamb.taxvamb_encode.make_dataloader_labels_hloss(
        abundance,
        tnfs,
        lengths,
        targets,
        len(nodes),
        table_parent,
        batchsize=vae_options.basic_options.starting_batch_size,
        cuda=opt.common.general.cuda,
    )

    shapes = (abundance.shape[1], 103, 1, len(nodes))
    dataloader = vamb.taxvamb_encode.make_dataloader_semisupervised_hloss(
        dataloader_joint,
        dataloader_vamb,
        dataloader_labels,
        len(nodes),
        table_parent,
        shapes,
        opt.common.general.seed,
        batchsize=vae_options.basic_options.starting_batch_size,
        cuda=opt.common.general.cuda,
    )
    model_path = opt.common.general.out_dir.joinpath("vaevae_model.pt")
    with open(model_path, "wb") as modelfile:
        vae.trainmodel(
            dataloader,
            nepochs=vae_options.basic_options.num_epochs,
            modelfile=modelfile,
            batchsteps=vae_options.basic_options.batch_steps,
        )

    latent_both = vae.VAEJoint.encode(dataloader_joint)
    logger.info(f"{latent_both.shape} embedding shape")

    latent_path = opt.common.general.out_dir.joinpath("vaevae_latent.npz")
    vamb.vambtools.write_npz(latent_path, latent_both)

    # Cluster, save tsv file
    cluster_and_write_files(
        opt.common.clustering,
        opt.common.output.binsplitter,
        latent_both,
        cast(Sequence[str], contignames),
        lengths,
        opt.common.general.seed,
        opt.common.general.cuda,
        str(opt.common.general.out_dir.joinpath("vaevae_clusters")),
        FastaOutput.try_from_common(opt.common),
        None,
    )


def run_reclustering(opt: ReclusteringOptions):
    composition = calc_tnf(
        opt.composition,
        opt.general.min_contig_length,
        opt.general.out_dir,
        opt.output.binsplitter,
    )
    markers = load_markers(
        opt.markers, composition.metadata, opt.general.out_dir, opt.general.n_threads
    )
    latent = vamb.vambtools.read_npz(opt.latent_path)
    alg = opt.algorithm

    if isinstance(alg, DBScanOptions):
        # If we should refine or not.
        if isinstance(alg.taxonomy, TaxometerOptions):
            taxopt = alg.taxonomy
            abundance = calc_abundance(
                taxopt.abundance,
                taxopt.general.out_dir,
                taxopt.general.refcheck,
                composition.metadata,
                taxopt.general.n_threads,
            )
            predicted_tax = predict_taxonomy(
                composition.metadata,
                abundance.matrix,
                composition.matrix,
                composition.metadata.lengths,
                taxopt.general.out_dir,
                taxopt,
                taxopt.general.cuda,
            )
            taxonomy = predicted_tax.to_taxonomy()
        else:
            if isinstance(alg.taxonomy, UnrefinedTaxonomy):
                logger.info(
                    f'Loading unrefined taxonomy from file "{alg.taxonomy.path}"'
                )
                taxonomy = vamb.taxonomy.Taxonomy.from_file(
                    alg.taxonomy.path, composition.metadata, True
                )
            else:
                logger.info(f'Loading refined taxonomy from file "{alg.taxonomy.path}"')
                taxonomy = vamb.taxonomy.Taxonomy.from_refined_file(
                    alg.taxonomy.path, composition.metadata, True
                )

        instantiated_alg = vamb.reclustering.DBScanAlgorithm(
            composition.metadata, taxonomy, opt.general.n_threads
        )
        logger.info("Reclustering")
        logger.info("\tAlgorithm: DBSCAN")
        reclustered_contigs = vamb.reclustering.recluster_bins(
            markers, latent, instantiated_alg
        )
    else:
        with open(alg.clusters) as file:
            clusters = vamb.vambtools.read_clusters(file)
        contig_to_id: dict[str, int] = {
            c: i for (i, c) in enumerate(composition.metadata.identifiers)
        }
        clusters_as_ids: list[set[vamb.reclustering.ContigId]] = []
        for cluster in clusters.values():
            s = set()
            for contig in cluster:
                i = contig_to_id.get(contig, None)
                if i is None:
                    raise ValueError(
                        f'Contig "{contig}" found in the provided clusters file is not found in the '
                        "provided composition."
                    )
                s.add(vamb.reclustering.ContigId(i))
            clusters_as_ids.append(s)
        instantiated_alg = vamb.reclustering.KmeansAlgorithm(
            clusters_as_ids,
            abs(opt.general.seed) % 4294967295,  # Note: sklearn seeds must be uint32
            composition.metadata.lengths,
        )
        logger.info("Reclustering")
        logger.info("\tAlgorithm: KMeans")
        reclustered_contigs = vamb.reclustering.recluster_bins(
            markers, latent, instantiated_alg
        )

    logger.info("\tReclustering complete")
    identifiers = composition.metadata.identifiers
    clusters_dict: dict[str, set[str]] = dict()
    for i, cluster in enumerate(reclustered_contigs):
        clusters_dict[str(i)] = {identifiers[c] for c in cluster}

    if opt.output.min_fasta_output_size is None:
        fasta_output = None
    else:
        assert isinstance(opt.composition.path, FASTAPath)
        fasta_output = FastaOutput(
            opt.composition.path,
            opt.general.out_dir.joinpath("bins"),
            opt.output.min_fasta_output_size,
        )

    write_clusters_and_bins(
        fasta_output,
        None,
        opt.output.binsplitter,
        str(opt.general.out_dir.joinpath("clusters_reclustered")),
        clusters_dict,
        cast(Sequence[str], composition.metadata.identifiers),
        cast(Sequence[int], composition.metadata.lengths),
    )


def add_help_arguments(parser: argparse.ArgumentParser):
    helpos = parser.add_argument_group(title="Help and version", description=None)
    helpos.add_argument("-h", "--help", help="Print help and exit", action="help")
    helpos.add_argument(
        "--version",
        action="version",
        version=f"Vamb {vamb.__version_str__}",
    )


def add_general_arguments(subparser: argparse.ArgumentParser):
    add_help_arguments(subparser)
    reqos = subparser.add_argument_group(title="Output", description=None)
    reqos.add_argument(
        "--outdir",
        metavar="",
        type=Path,
        # required=True, # Setting this flag breaks the argparse with positional arguments. The existence of the folder is checked elsewhere
        help="Output directory to create",
        required=True,
    )

    general = subparser.add_argument_group(
        title="General optional arguments", description=None
    )
    general.add_argument(
        "-m",
        dest="minlength",
        metavar="",
        type=int,
        default=2000,
        help="Ignore contigs shorter than this [2000]",
    )

    general.add_argument(
        "-p",
        dest="nthreads",
        metavar="",
        type=int,
        default=DEFAULT_THREADS,
        help=f"number of threads to use where customizable [{DEFAULT_THREADS}]",
    )
    general.add_argument(
        "--norefcheck",
        help="Skip reference name hashing check [False]",
        action="store_true",
    )
    general.add_argument(
        "--cuda", help="Use GPU to train & cluster [False]", action="store_true"
    )
    general.add_argument(
        "--seed",
        metavar="",
        type=int,
        default=int.from_bytes(os.urandom(7), "little"),
        help="Random seed (determinism not guaranteed)",
    )
    return subparser


def add_composition_arguments(subparser: argparse.ArgumentParser):
    tnfos = subparser.add_argument_group(title="Composition input")
    tnfos.add_argument("--fasta", metavar="", type=Path, help="Path to fasta file")
    tnfos.add_argument(
        "--composition", metavar="", type=Path, help="Path to .npz of composition"
    )
    return subparser


def add_abundance_arguments(subparser: argparse.ArgumentParser):
    abundanceos = subparser.add_argument_group(title="Abundance input")
    # Note: This argument is deprecated, but we'll keep supporting it for now.
    # Instead, use --bamdir.
    abundanceos.add_argument(
        "--bamfiles",
        dest="bampaths",
        metavar="",
        type=Path,
        help=argparse.SUPPRESS,
        nargs="+",
    )
    abundanceos.add_argument(
        "--bamdir",
        metavar="",
        type=Path,
        help="Dir with .bam files to use",
    )
    abundanceos.add_argument(
        "--abundance_tsv",
        metavar="",
        type=Path,
        help='Path to TSV file of precomputed abundances with header being "contigname(\\t<samplename>)*"',
    )
    abundanceos.add_argument(
        "--abundance",
        metavar="",
        dest="abundancepath",
        type=Path,
        help="Path to .npz of abundances",
    )
    abundanceos.add_argument(
        "-z",
        dest="min_alignment_id",
        metavar="",
        type=float,
        default=None,
        help=argparse.SUPPRESS,
    )
    return subparser


def add_taxonomy_arguments(subparser: argparse.ArgumentParser, taxonomy_only=False):
    taxonomys = subparser.add_argument_group(title="Taxonomy input")
    taxonomys.add_argument(
        "--taxonomy",
        metavar="",
        type=Path,
        help="Path to the taxonomy file",
    )
    if not taxonomy_only:
        taxonomys.add_argument(
            "--no_predictor",
            help="Do not complete input taxonomy with Taxometer [False]",
            action="store_true",
        )
    return subparser


def add_marker_arguments(subparser: argparse.ArgumentParser):
    marker_s = subparser.add_argument_group(title="Marker gene input")
    marker_s.add_argument(
        "--markers",
        metavar="",
        type=Path,
        help="Path to the marker .npz file",
    )
    marker_s.add_argument(
        "--hmm_path",
        metavar="",
        type=Path,
        help="Path to the .hmm file of marker gene profiles",
    )
    return subparser


def add_bin_output_arguments(subparser: argparse.ArgumentParser):
    bin_os = subparser.add_argument_group(title="Bin output options", description=None)
    bin_os.add_argument(
        "--minfasta",
        dest="min_fasta_output_size",
        metavar="",
        type=int,
        default=None,
        help="Minimum bin size to output as fasta [None = no files]",
    )
    bin_os.add_argument(
        "-o",
        dest="binsplit_separator",
        metavar="",
        type=str,
        # This means: None is not passed, "" if passed but empty, e.g. "-o -c 5"
        # otherwise a string.
        default=None,
        const="",
        nargs="?",
        help="Binsplit separator [C if present] (pass empty string to disable)",
    )
    return subparser


def add_vae_arguments(subparser: argparse.ArgumentParser):
    # VAE arguments
    vaeos = subparser.add_argument_group(title="VAE options", description=None)

    vaeos.add_argument(
        "-n",
        dest="nhiddens",
        metavar="",
        type=int,
        nargs="+",
        default=None,
        help=argparse.SUPPRESS,
    )
    vaeos.add_argument(
        "-l",
        dest="nlatent",
        metavar="",
        type=int,
        default=32,
        help=argparse.SUPPRESS,
    )
    vaeos.add_argument(
        "-a",
        dest="alpha",
        metavar="",
        type=float,
        default=None,
        help=argparse.SUPPRESS,
    )
    vaeos.add_argument(
        "-b",
        dest="beta",
        metavar="",
        type=float,
        default=200.0,
        help=argparse.SUPPRESS,
    )
    vaeos.add_argument(
        "-d",
        dest="dropout",
        metavar="",
        type=float,
        default=None,
        help=argparse.SUPPRESS,
    )

    trainos = subparser.add_argument_group(title="Training options", description=None)

    trainos.add_argument(
        "-e", dest="nepochs", metavar="", type=int, default=300, help=argparse.SUPPRESS
    )
    trainos.add_argument(
        "-t",
        dest="batchsize",
        metavar="",
        type=int,
        default=256,
        help=argparse.SUPPRESS,
    )
    trainos.add_argument(
        "-q",
        dest="batchsteps",
        metavar="",
        type=int,
        nargs="*",
        default=[25, 75, 150, 225],
        help=argparse.SUPPRESS,
    )
    trainos.add_argument(
        "-r",
        dest="lrate",
        metavar="",
        type=float,
        default=None,
        help=argparse.SUPPRESS,
    )
    return subparser


def add_predictor_arguments(subparser: argparse.ArgumentParser):
    # Train predictor arguments
    pred_trainos = subparser.add_argument_group(
        title="Training options for the taxonomy predictor", description=None
    )

    pred_trainos.add_argument(
        "-pe",
        dest="pred_nepochs",
        metavar="",
        type=int,
        default=100,
        help=argparse.SUPPRESS,
    )
    pred_trainos.add_argument(
        "-pt",
        dest="pred_batchsize",
        metavar="",
        type=int,
        default=1024,
        help=argparse.SUPPRESS,
    )
    pred_trainos.add_argument(
        "-pq",
        dest="pred_batchsteps",
        metavar="",
        type=int,
        nargs="*",
        default=[25, 75],
        help=argparse.SUPPRESS,
    )
    pred_trainos.add_argument(
        "-pthr",
        dest="pred_softmax_threshold",
        metavar="",
        type=float,
        default=0.5,
        help=argparse.SUPPRESS,
    )
    pred_trainos.add_argument(
        "-ploss",
        dest="ploss",
        metavar="",
        type=str,
        choices=["flat_softmax", "cond_softmax", "soft_margin"],
        default="flat_softmax",
        help=argparse.SUPPRESS,
    )
    return subparser


def add_clustering_arguments(subparser: argparse.ArgumentParser):
    # Clustering arguments
    clusto = subparser.add_argument_group(title="Clustering options", description=None)
    clusto.add_argument(
        "-w",
        dest="window_size",
        metavar="",
        type=int,
        default=300,
        help=argparse.SUPPRESS,
    )
    clusto.add_argument(
        "-u",
        dest="min_successes",
        metavar="",
        type=int,
        default=15,
        help=argparse.SUPPRESS,
    )
    clusto.add_argument(
        "-c",
        dest="max_clusters",
        metavar="",
        type=int,
        default=None,  # meaning: do not stop
        help=argparse.SUPPRESS,
    )

    return subparser


def add_model_selection_arguments(subparser: argparse.ArgumentParser):
    # Model selection argument
    model_selection = subparser.add_argument_group(
        title="Model selection", description=None
    )

    model_selection.add_argument(
        "--model",
        dest="model",
        metavar="",
        type=str,
        choices=[
            "vae",
            "aae",
            "vae-aae",
            "vaevae",
        ],
        default="vae",
        help="Choose which model to run; only vae (vae); only aae (aae); the combination of vae and aae (vae-aae); taxonomy_predictor; vaevae; reclustering [vae]",
    )
    return subparser


def add_aae_arguments(subparser: argparse.ArgumentParser):
    # AAE arguments
    aaeos = subparser.add_argument_group(title="AAE options", description=None)

    aaeos.add_argument(
        "--n_aae",
        dest="nhiddens_aae",
        metavar="",
        type=int,
        default=547,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--z_aae",
        dest="nlatent_aae_z",
        metavar="",
        type=int,
        default=283,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--y_aae",
        dest="nlatent_aae_y",
        metavar="",
        type=int,
        default=700,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--sl_aae",
        dest="sl",
        metavar="",
        type=float,
        default=0.00964,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--slr_aae",
        dest="slr",
        metavar="",
        type=float,
        default=0.5,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--aae_temp",
        dest="temp",
        metavar="",
        type=float,
        default=0.1596,
        help=argparse.SUPPRESS,
    )

    aaeos.add_argument(
        "--e_aae",
        dest="nepochs_aae",
        metavar="",
        type=int,
        default=70,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--t_aae",
        dest="batchsize_aae",
        metavar="",
        type=int,
        default=256,
        help=argparse.SUPPRESS,
    )
    aaeos.add_argument(
        "--q_aae",
        dest="batchsteps_aae",
        metavar="",
        type=int,
        nargs="*",
        default=[25, 50],
        help=argparse.SUPPRESS,
    )
    return subparser


def add_reclustering_arguments(subparser: argparse.ArgumentParser):
    # Reclustering arguments
    reclusters = subparser.add_argument_group(title="K-means reclustering arguments")
    reclusters.add_argument(
        "--latent_path",
        metavar="",
        type=Path,
        help="Path to latent space .npz file",
    )
    reclusters.add_argument(
        "--clusters_path",
        metavar="",
        type=Path,
        help="Path to TSV file with clusters",
    )
    reclusters.add_argument(
        "--algorithm",
        metavar="",
        type=str,
        default="kmeans",
        choices=["kmeans", "dbscan"],
        help="Which reclustering algorithm to use ('kmeans', 'dbscan') [kmeans]",
    )
    return subparser


def main():
    doc = f"""
    Version: {vamb.__version_str__}

    Default use, good for most datasets:
    vamb bin default --outdir out --fasta my_contigs.fna --bamdir bam_dir

    Find the latest updates and documentation at https://github.com/RasmussenLab/vamb"""
    logger.add(sys.stderr, format=format_log)

    parser = argparse.ArgumentParser(
        prog="vamb",
        description=doc,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False,
    )
    add_help_arguments(parser)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()

    subparsers = parser.add_subparsers(dest="subcommand")

    TAXOMETER = "taxometer"
    BIN = "bin"
    VAMB = "default"
    TAXVAMB = "taxvamb"
    AVAMB = "avamb"
    RECLUSTER = "recluster"

    vaevae_parserbin_parser = subparsers.add_parser(
        BIN,
        help="""
        VAMB, TaxVAMB, AVAMB binners
        """,
        add_help=False,
    )
    add_help_arguments(vaevae_parserbin_parser)
    subparsers_model = vaevae_parserbin_parser.add_subparsers(dest="model_subcommand")

    vae_parser = subparsers_model.add_parser(
        VAMB,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="""
        default binner based on a variational autoencoder. 
        See the paper "Improved metagenome binning and assembly using deep variational autoencoders" 
        (https://www.nature.com/articles/s41587-020-00777-4)
        """,
        add_help=False,
        usage="%(prog)s [options]",
        description="""Bin using a VAE that merges composition and abundance information.

Required arguments: Outdir, at least one composition input and at least one abundance input""",
    )
    add_general_arguments(vae_parser)
    add_composition_arguments(vae_parser)
    add_abundance_arguments(vae_parser)
    add_bin_output_arguments(vae_parser)
    add_vae_arguments(vae_parser)
    add_clustering_arguments(vae_parser)

    vaevae_parser = subparsers_model.add_parser(
        TAXVAMB,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="""
        taxonomy informed binner based on a bi-modal variational autoencoder. 
        See the paper "TaxVAMB: taxonomic annotations improve metagenome binning" 
        (link)
        """,
        add_help=False,
        usage="%(prog)s [options]",
        description="""Bin using a semi-supervised VAEVAE model that merges composition, abundance and taxonomic information.

Required arguments: Outdir, taxonomy, at least one composition input and at least one abundance input""",
    )
    add_general_arguments(vaevae_parser)
    add_composition_arguments(vaevae_parser)
    add_abundance_arguments(vaevae_parser)
    add_taxonomy_arguments(vaevae_parser)
    add_bin_output_arguments(vaevae_parser)
    add_vae_arguments(vaevae_parser)
    add_clustering_arguments(vaevae_parser)
    add_predictor_arguments(vaevae_parser)

    vaeaae_parser = subparsers_model.add_parser(
        AVAMB,
        help=argparse.SUPPRESS,
        add_help=False,
        usage="%(prog)s [options]",
    )
    add_general_arguments(vaeaae_parser)
    add_composition_arguments(vaeaae_parser)
    add_abundance_arguments(vaeaae_parser)
    add_bin_output_arguments(vaeaae_parser)
    add_vae_arguments(vaeaae_parser)
    add_aae_arguments(vaeaae_parser)
    add_clustering_arguments(vaeaae_parser)

    predict_parser = subparsers.add_parser(
        TAXOMETER,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="""
        refines taxonomic annotations of any metagenome classifier. 
        See the paper "Taxometer: deep learning improves taxonomic classification of contigs using binning features and a hierarchical loss" 
        (link)
        """,
        add_help=False,
        usage="%(prog)s [options]",
        description="""Refine taxonomy using composition and abundance information.

Required arguments: Outdir, unrefined taxonomy, at least one composition input and at least one abundance input""",
    )
    add_general_arguments(predict_parser)
    add_composition_arguments(predict_parser)
    add_abundance_arguments(predict_parser)
    add_taxonomy_arguments(predict_parser, taxonomy_only=True)
    add_predictor_arguments(predict_parser)

    recluster_parser = subparsers.add_parser(
        RECLUSTER,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="""
        reclustering using single-copy genes for the binning results of VAMB, TaxVAMB or AVAMB
        """,
        add_help=False,
        usage="%(prog)s [options]",
        description="""Use marker genes to re-cluster (DBScan) or refine (K-means) clusters.
        
Required arguments:
  K-means algorithm: Outdir, at least one composition input, at least one marker gene input,
    latent path and clusters path
  DBScan algorithm with --no_predictor: Outdir, at least one composition input,
    at least one marker gene input, latent path and taxonomy
  DBScan algorithm without --no_predictor: Outdir, at least one composition input,
    at least one abundance input, at least one marker gene input, latent path and taxonomy
""",
    )
    add_general_arguments(recluster_parser)
    add_composition_arguments(recluster_parser)
    add_abundance_arguments(recluster_parser)
    add_marker_arguments(recluster_parser)
    add_bin_output_arguments(recluster_parser)
    add_reclustering_arguments(recluster_parser)
    add_predictor_arguments(recluster_parser)
    add_taxonomy_arguments(recluster_parser)

    args = parser.parse_args()

    if args.subcommand == TAXOMETER:
        opt = TaxometerOptions.from_args(args)
        runner = partial(run_taxonomy_predictor, opt)
        run(runner, opt.general)
    elif args.subcommand == BIN:
        model = args.model_subcommand
        if model is None:
            vaevae_parserbin_parser.print_help()
            sys.exit(1)
        if model == VAMB:
            opt = BinDefaultOptions.from_args(args)
            runner = partial(run_bin_default, opt)
            run(runner, opt.common.general)
        elif model == TAXVAMB:
            opt = BinTaxVambOptions.from_args(args)
            runner = partial(run_vaevae, opt)
            run(runner, opt.common.general)
        elif model == AVAMB:
            opt = BinAvambOptions.from_args(args)
            runner = partial(run_bin_aae, opt)
            run(runner, opt.common.general)
    elif args.subcommand == RECLUSTER:
        opt = ReclusteringOptions.from_args(args)
        runner = partial(run_reclustering, opt)
        run(runner, opt.general)
    else:
        # There are no more subcommands
        assert False


if __name__ == "__main__":
    main()



import unittest
import io
import random
import numpy as np
import torch
from hashlib import sha256

import vamb
import testtools

# PyTorch cannot be made stable, so we cannot run CI on the result
# of pytorch training.
# Nonetheless, you can enable it locally with this switch here,
# which can be useful sometimes.
TEST_UNSTABLE_HASHES = False


class TestCompositionResult(unittest.TestCase):
    io = io.BytesIO()

    @classmethod
    def setUpClass(cls):
        rng = random.Random(15)
        for _ in range(4):
            seq = testtools.make_randseq(rng, 400, 600)
            cls.io.write(seq.format().encode())
            cls.io.write(b"\n")

    def setUp(self):
        self.io.seek(0)

    def test_runs(self):
        comp = vamb.parsecontigs.Composition.from_file(self.io)
        self.assertIsInstance(comp, vamb.parsecontigs.Composition)

    if TEST_UNSTABLE_HASHES:

        def test_result(self):
            comp = vamb.parsecontigs.Composition.from_file(self.io)
            self.assertEqual(
                sha256(comp.matrix.data.tobytes()).digest().hex(),
                "9e9a2d7b021654e874894722bdd6cd3eda18bed03fabd32a9440e806a8ab1bd1",
            )


class TestAbundanceResult(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.comp_metadata = vamb.parsecontigs.CompositionMetaData(
            np.array(testtools.BAM_NAMES, dtype=object),
            np.array(testtools.BAM_SEQ_LENS),
            np.ones(len(testtools.BAM_SEQ_LENS), dtype=bool),
            2000,
        )

    def test_runs(self):
        abundance = vamb.parsebam.Abundance.from_files(
            testtools.BAM_FILES, None, self.comp_metadata, True, 0.9, 4
        )
        self.assertIsInstance(abundance, vamb.parsebam.Abundance)

    if TEST_UNSTABLE_HASHES:

        def test_result(self):
            abundance = vamb.parsebam.Abundance.from_files(
                testtools.BAM_FILES, "/tmp/tmpbam", self.comp_metadata, True, 0.9, 2
            )
            self.assertEqual(
                sha256(abundance.matrix.data.tobytes()).digest().hex(),
                "c346abb53b62423fe95ed4b2eb5988d77141b2d7a5c58c03fdf09abc6476df78",
            )
            abundance2 = vamb.parsebam.Abundance.from_files(
                testtools.BAM_FILES, None, self.comp_metadata, True, 0.9, 4
            )
            self.assertTrue(np.all(np.abs(abundance.matrix - abundance2.matrix) < 1e-5))


class TestEncodingResult(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        torch.manual_seed(0)
        rng = np.random.RandomState(15)
        cls.tnfs = rng.random((200, 103)).astype(np.float32)
        cls.rpkm = rng.random((200, 6)).astype(np.float32)
        cls.lens = rng.randint(2000, 5000, 200)

    def test_runs(self):
        self.assertEqual(
            sha256(self.lens.data.tobytes()).digest().hex(),
            "68894f01cc435a5f032a655faecddd817cd35a71397129296a11f8c40bd29fcb",
        )

        vae = vamb.encode.VAE(6)
        dl = vamb.encode.make_dataloader(
            self.rpkm.copy(), self.tnfs, self.lens, batchsize=16
        )
        vae.trainmodel(dl, nepochs=3, batchsteps=[1, 2])
        latent = vae.encode(dl)

        self.assertIsInstance(latent, np.ndarray)

    if TEST_UNSTABLE_HASHES:

        def test_result(self):
            torch.manual_seed(0)
            torch.use_deterministic_algorithms(True)
            np.random.seed(0)
            random.seed(0)
            vae = vamb.encode.VAE(6)
            dl = vamb.encode.make_dataloader(
                self.rpkm, self.tnfs, self.lens, batchsize=16
            )
            vae.trainmodel(dl, nepochs=3, batchsteps=[1, 2])
            latent = vae.encode(dl)

            self.assertEqual(
                sha256(latent.data.tobytes()).digest().hex(),
                "0148ec0767e88c756615340d6fd0b31ca07aa6b4b172a1874fb7de7179acb57d",
            )

            self.assertEqual(
                sha256(torch.rand(10).numpy().tobytes()).digest().hex(),
                "c417b9722e14e854fbe79cc5c797cc6653360c1e6536064205ca0c073f41eaf6",
            )


class TestClusterResult(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        rng = np.random.RandomState(15)
        cls.latent = rng.random((1000, 3)).astype(np.float32) - 0.5

    def test_runs(self):
        self.assertEqual(
            sha256(self.latent.tobytes()).digest().hex(),
            "630a98a4b44c3754a3f423e915847f44767bb69fb13ea5901dc512428aee9811",
        )

    if TEST_UNSTABLE_HASHES:

        def test_result(self):
            hash = sha256()

            # Use this to check that the clustering used in this test produces
            # a reasonable cluster size, and that it doesn't just pass because
            # it always clusters everything in 1-point clusters.
            # Uncomment when updating this test.
            # lens = list()
            for cluster in vamb.cluster.ClusterGenerator(self.latent.copy()):
                medoid = cluster.metadata.medoid
                points = set(cluster.members)
                # Set hashing may differ from run to run, so turn into sorted arrays
                arr = np.array(list(points))
                arr.sort()
                # lens.append(arr)
                hash.update(medoid.to_bytes(4, "big"))
                hash.update(arr.data)

            # self.assertGreater(len(list(map(lambda x: len(lens) > 1))), 3)
            self.assertEqual(
                hash.digest().hex(),
                "2b3caf674ff1d1906a831219e0953b2d9f1b78ecefec709b70c672280af49aee",
            )
